# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2019 Osyris contributors (https://github.com/nvaytet/osyris)
# @author Neil Vaytet

from .engine import get_binary_data
from . import config as conf
import numpy as np
from scipy.interpolate import RegularGridInterpolator

#===================================================================================
# An empty container class which will be filled up as the EOS, opacity, resistivity
# tables are read.
#===================================================================================
class IsmTable():

    def __init__(self):

        return


#===================================================================================
# Generic function to interpolate simulation data onto opacity table
#===================================================================================
def ism_interpolate(table_container=None,values=[0],points=[0],in_log=False):

    func = RegularGridInterpolator(table_container.grid,values)

    if in_log:
        return func(points)
    else:
        return np.power(10.0,func(points))


#===================================================================================
# Function to read in binary file containing EOS table
#===================================================================================
def read_eos_table(fname="tab_eos.dat"):

    print("Loading EOS table: "+fname)

    # Read binary EOS file
    with open(fname, mode='rb') as eos_file:
        eosContent = eos_file.read()
    eos_file.close()

    # Define data fields. Note that the order is important!
    data_fields = ["rho_eos","ener_eos","temp_eos","pres_eos","s_eos","cs_eos","xH_eos","xH2_eos","xHe_eos","xHep_eos"]

    # Create table container
    theTable = IsmTable()

    # Initialise offset counters and start reading data
    ninteg = nfloat = nlines = nstrin = nquadr = 0

    # Get table dimensions
    theTable.nx = np.array(get_binary_data(fmt="2i",content=eosContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat))

    # Get table limits
    ninteg += 2
    nlines += 1
    [theTable.rhomin,theTable.rhomax,theTable.emin,theTable.emax,theTable.yHe] = \
        get_binary_data(fmt="5d",content=eosContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    array_size = np.prod(theTable.nx)
    array_fmt  = "%id" % array_size
    nfloat += 5
    nlines += 1

    # Now loop through all the data fields
    for i in range(len(data_fields)):
        setattr(theTable,data_fields[i],np.reshape(get_binary_data(fmt=array_fmt,content=eosContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F"))
        nfloat += array_size
        nlines += 1

    del eosContent

    #theTable.grid = (np.log10(theTable.rho_eos[:,0]), np.log10(theTable.ener_eos[0,:]/theTable.rho_eos[0,:]))

    Eint = theTable.ener_eos/theTable.rho_eos
    theTable.grid = (np.log10(theTable.rho_eos[:,0]), np.log10(Eint[0,:]))

    #print theTable.rho_eos[:,0]
    #print Eint[0,:]
    #print theTable.ener_eos[0,:]

    print("EOS table read successfully")

    return theTable


#===================================================================================
# Function to interpolate simulation data onto EOS table
#===================================================================================
def get_eos(holder,fname="tab_eos.dat",variables=["temp_eos","pres_eos","s_eos","cs_eos","xH_eos","xH2_eos","xHe_eos","xHep_eos"]):

    if holder.info["eos"] == 0:
        print("Simulation data did not use a tabulated EOS. Exiting")
        return
    else:
        try:
            n = holder.eos_table.nRho
        except AttributeError:
            holder.eos_table = read_eos_table(fname=fname)

        pts = np.array([np.log10(holder.density.values), np.log10(holder.internal_energy.values/holder.density.values)]).T
        for var in variables:
            print("Interpolating "+var)
            vals = ism_interpolate(holder.eos_table,np.log10(getattr(holder.eos_table,var)),pts)
            holder.new_field(name=var,label=var,values=vals,verbose=False)

    return


#===================================================================================
#===================================================================================
#===================================================================================

#===================================================================================
# Function to read in binary file containing opacity table
#===================================================================================
def read_opacity_table(fname="vaytet_grey_opacities3D.bin"):

    print("Loading opacity table: "+fname)

    # Read binary resistivity file
    with open(fname, mode='rb') as kappa_file:
        kappaContent = kappa_file.read()
    kappa_file.close()

    # Create table container
    theTable = IsmTable()

    # Initialise offset counters and start reading data
    ninteg = nfloat = nlines = nstrin = nquadr = 0

    # Get table dimensions
    theTable.nx = np.array(get_binary_data(fmt="3i",content=kappaContent))

    ## Get table limits
    #[theTable.dx,theTable.dy,theTable.dz,theTable.xmin,theTable.xmax,theTable.ymin,theTable.ymax,theTable.zmin,theTable.zmax] = \
        #get_binary_data(fmt="9d",content=kappaContent,correction=12)

    # Read table coordinates:

    # x: density
    ninteg += 3
    nfloat += 9
    nlines += 1
    theTable.dens = get_binary_data(fmt="%id"%theTable.nx[0],content=kappaContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    # y: gas temperature
    nfloat += theTable.nx[0]
    nlines += 1
    theTable.tgas = get_binary_data(fmt="%id"%theTable.nx[1],content=kappaContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    # z: radiation temperature
    nfloat += theTable.nx[1]
    nlines += 1
    theTable.trad = get_binary_data(fmt="%id"%theTable.nx[2],content=kappaContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    # Now read opacities
    array_size = np.prod(theTable.nx)
    array_fmt  = "%id" % array_size

    #print theTable.nx,theTable.ny,theTable.nz

    # Planck mean
    nfloat += theTable.nx[2]
    nlines += 1
    theTable.kappa_p = np.reshape(get_binary_data(fmt=array_fmt,content=kappaContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F")

    # Rosseland mean
    nfloat += array_size
    nlines += 1
    theTable.kappa_r = np.reshape(get_binary_data(fmt=array_fmt,content=kappaContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F")

    del kappaContent

    theTable.grid = (theTable.dens,theTable.tgas,theTable.trad)

    #print theTable.dens
    #print np.shape(theTable.dens)
    ##print theTable.tgas
    ##print theTable.trad


    print("Opacity table read successfully")

    return theTable


#===================================================================================
# Function to interpolate simulation data onto opacity table
#===================================================================================
def get_opacities(holder,fname="vaytet_grey_opacities3D.bin",variables=["kappa_p","kappa_r"]):

    try:
        n = holder.opacity_table.nx
    except AttributeError:
        holder.opacity_table = read_opacity_table(fname=fname)

    if not hasattr(holder,"radiative_temperature"):
        print("Radiative temperature is not defined. Computing it now.")
        holder.new_field(name="radiative_temperature",operation="(radiative_energy_1/"+str(conf.constants["a_r"])+")**0.25",label="Trad")

    pts = np.array([np.log10(holder.density.values),np.log10(holder.temperature.values),np.log10(holder.radiative_temperature.values)]).T
    for var in variables:
        print("Interpolating "+var)
        vals = ism_interpolate(holder.opacity_table,getattr(holder.opacity_table,var),pts)
        holder.new_field(name=var,label=var,values=vals,verbose=False,unit="cm2/g")

    return


#===================================================================================
#===================================================================================
#===================================================================================

#===================================================================================
# Function to read in binary file containing resistivity table
#===================================================================================
def read_resistivity_table(fname="resistivities_masson2016.bin"):

    print("Loading resistivity table: "+fname)

    # Read binary resistivity file
    with open(fname, mode='rb') as res_file:
        resContent = res_file.read()
    res_file.close()

    # Create table container
    theTable = IsmTable()

    # Initialise offset counters and start reading data
    ninteg = nfloat = nlines = nstrin = nquadr = 0

    # Get length of record on first line to determine number of dimensions in table
    rec_size = get_binary_data(fmt="i",content=resContent,correction=-4)
    theTable.ndims = int(rec_size[0]/4)

    # Get table dimensions
    theTable.nx = np.array(get_binary_data(fmt="%ii"%theTable.ndims,content=resContent))

    # Read table coordinates:

    # 1: density
    ninteg += theTable.ndims
    nlines += 1
    theTable.dens = get_binary_data(fmt="%id"%theTable.nx[0],content=resContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    # 2: gas temperature
    nfloat += theTable.nx[0]
    nlines += 1
    theTable.tgas = get_binary_data(fmt="%id"%theTable.nx[1],content=resContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    if theTable.ndims == 4:
        # 3: ionisation rate
        nfloat += theTable.nx[1]
        nlines += 1
        theTable.ionx = get_binary_data(fmt="%id"%theTable.nx[2],content=resContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    # 4: magnetic field
    nfloat += theTable.nx[-2]
    nlines += 1
    theTable.bmag = get_binary_data(fmt="%id"%theTable.nx[-1],content=resContent,ninteg=ninteg,nlines=nlines,nfloat=nfloat)

    # Now read resistivities
    array_size = np.prod(theTable.nx)
    array_fmt  = "%id" % array_size

    # Ohmic resistivity
    nfloat += theTable.nx[-1]
    nlines += 1
    theTable.eta_ohm = np.reshape(get_binary_data(fmt=array_fmt,content=resContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F")

    # Ambipolar resistivity
    nfloat += array_size
    nlines += 1
    theTable.eta_ad = np.reshape(get_binary_data(fmt=array_fmt,content=resContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F")

    # Hall resistivity
    nfloat += array_size
    nlines += 1
    theTable.eta_hall = np.reshape(get_binary_data(fmt=array_fmt,content=resContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F")

    # Hall sign
    nfloat += array_size
    nlines += 1
    theTable.eta_hsig = np.reshape(get_binary_data(fmt=array_fmt,content=resContent, \
                ninteg=ninteg,nlines=nlines,nfloat=nfloat),theTable.nx,order="F")

    del resContent

    if theTable.ndims == 4:
        theTable.grid = (theTable.dens,theTable.tgas,theTable.ionx,theTable.bmag)
    elif theTable.ndims == 3:
        theTable.grid = (theTable.dens,theTable.tgas,theTable.bmag)

    # Additional parameters
    theTable.scale_dens = 0.844*2.0/1.66e-24 # 2.0*H2_fraction/mH
    theTable.ionis_rate = 1.0e-17

    print("Resistivity table read successfully")

    return theTable


#===================================================================================
# Function to interpolate simulation data onto resistivity table
#===================================================================================
def get_resistivities(holder,fname="resistivities_masson2016.bin",variables=["eta_ohm","eta_ad","eta_hall"]):

    try:
        n = holder.resistivity_table.nx
    except AttributeError:
        holder.resistivity_table = read_resistivity_table(fname=fname)

    try:
        rho_to_nH = holder.resistivity_table.scale_dens/holder.info["mu_gas"]
    except KeyError:
        rho_to_nH = holder.resistivity_table.scale_dens/2.31 # Default mean atomic weight

    if holder.resistivity_table.ndims == 4:
        if not hasattr(holder,"ionisation_rate"):
            holder.new_field(name="ionisation_rate",values=[holder.resistivity_table.ionis_rate]*holder.info["ncells"],label="Ionisation rate")
        pts = np.array([np.log10(holder.density.values*rho_to_nH),np.log10(holder.temperature.values), \
                        np.log10(holder.ionisation_rate.values),np.log10(holder.B.values)]).T
    elif holder.resistivity_table.ndims == 3:
        pts = np.array([np.log10(holder.density.values*rho_to_nH),np.log10(holder.temperature.values), \
                        np.log10(holder.B.values)]).T

    for var in variables:
        print("Interpolating "+var)
        vals = ism_interpolate(holder.resistivity_table,getattr(holder.resistivity_table,var),pts)
        if var == "eta_hall":
            hall_sign = np.sign(ism_interpolate(holder.resistivity_table,holder.resistivity_table.eta_hsig,pts,in_log=True))
            holder.new_field(name=var,label=var,values=vals*hall_sign,verbose=False,unit="s")
        else:
            holder.new_field(name=var,label=var,values=vals,verbose=False,unit="s")

    return



from .load_ramses import RamsesData
from .plot import plot_histogram, plot_slice, plot_column_density
from . import ism_physics
from . import config
from .vtk import to_vtk
from .interpolate import interpolate



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2019 Osyris contributors (https://github.com/nvaytet/osyris)
# @author Neil Vaytet

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.collections
from matplotlib.colors import LogNorm, Normalize
from scipy.interpolate import griddata
from . import config as conf


def plot_histogram(var_x,
                   var_y,
                   axes=None,
                   block=False,
                   cbar=True,
                   clear=True,
                   contour=False,
                   contour_args={},
                   copy=False,
                   equal_axes=False,
                   fname=None,
                   image=False,
                   image_args={},
                   new_window=False,
                   only_leafs=True,
                   outline=False,
                   outline_args={},
                   plot=True,
                   resolution=256,
                   scalar=False,
                   scalar_args={},
                   scatter=False,
                   scatter_args={},
                   summed=False,
                   title=None,
                   update=None,
                   xmin=None,
                   xmax=None,
                   ymin=None,
                   ymax=None):
    """
    Plot a 2D histogram with two variables as input.
    This is used for instance to plot the temperature as a function of density
    for every cell in the mesh. You can supply a third variable for the
    colours. By default, if no variable for colour is supplied for either
    scalar, scatter, contour or image, the colour is used for the number of
    cells in each pixels.

    :param OsyrisField var_x: The variable for the x axis, e.g. mydata.log_rho.
    :param OsyrisField var_y: The variable for the x axis, e.g. mydata.log_T.
    :param MPL_axes axes: If specified, the data is plotted on the specified
        axes. The default is `None`.
    :param bool block: If ``True``, plotting the figure holds the terminal
        console until it is closed. The default is `False`.
    :param bool cbar: Make colorbar next to the plot if ``True``. The default
        is ``True``.
    :param bool clear: If ``True`` the figure is cleared before plotting. The
        default is `True`.
    :param OsyrisField contour: An osyris data field containing the variable
        to be represented using line contours. The default is ``False``.
    :param dict contour_args: A dictionary to hold additional matplotlib
        keyword arguments to be passed to the ``contour`` function. The default
        is empty.
    :param bool copy: If `True`, the function returns the data that was used to
        plot the histogram. Default is ``False``.

    """

    # :param equal_axes=False,
    # :param fname=None,
    # :param image=False,
    # :param image_args={},
    # :param new_window=False,
    # :param only_leafs=True,
    # :param outline=False,
    # :param outline_args={},
    # :param plot=True,
    # :param resolution=256,
    # :param scalar=False,
    # :param scalar_args={},
    # :param scatter=False,
    # :param scatter_args={},
    # :param summed=False,
    # :param title=None,
    # :param update=None,
    # :param xmin=None,
    # :param xmax=None,
    # :param ymin=None,
    # :param ymax=None):

    # :param str sender: The person sending the message
    # :param str recipient: The recipient of the message
    # :param str message_body: The body of the message
    # :param priority: The priority of the message, can be a number 1-5
    # :type priority: integer or None
    # :return: the message id
    # :rtype: int
    # :raises ValueError: if the message_body exceeds 160 characters
    # :raises TypeError: if the message_body is not a basestring

    # Parameters
    # ----------
    # x : type
    #     Description of parameter `x`.

    # var_x : osyrisField
    #     An osyris data field containing the variable for the x axis,
    #     e.g. mydata.log_rho.

    # var_y`: (*osyrisField*) An osyris data field containing the variable for the y
    # axis, e.g. mydata.log_T. There is no default.

    # scalar`: (*osyrisField*) An osyris data field containing the variable for the
    # colors to be used in the filled contours. Default is `False`.

    # image`: (*osyrisField*) An osyris data field containing the variable for the
    # colors to be used by the imshow function. Default is `False`.

    # contour`: (*osyrisField*) An osyris data field containing the variable for the
    # colors to be used in the filled contours. Default is `False`.

    # scatter`: (*osyrisField*) An osyris data field containing the variable for the
    # colors to be used in the scatter function. Default is `False`. **Note:**

    # fname`: (*string*) If specified, the figure is saved to file. Default is `None`.

    # axes`: (*MPL axes*) If specified, the data is plotted on the specified axes (see
    # demo). Default is `None`.

    # resolution`: (*integer*) The data is binned in a 2D matrix of size `resolution`.
    # Default is 256.

    # copy`: (*logical*) If `True`, the function returns the data that was used to plot
    # the histogram. Default is `False`.

    # title`: (*string*) The title for the figure/axes. Default is `None`. If no title
    # is specified, the simulation time will be used as a title. If you want no title on
    # the figure, use `title=""`.

    # xmin`: (*float*) Lower limit for the x axis. Default is `None`, implying an
    # automatic search for the lowest x value in the supplied var_x data.

    # xmax`: (*float*) Upper limit for the x axis. Default is `None`, implying an
    # automatic search for the highest x value in the supplied var_x data.

    # ymin`: (*float*) Lower limit for the y axis. Default is `None`, implying an
    # automatic search for the lowest y value in the supplied var_x data.

    # ymax`: (*float*) Upper limit for the y axis. Default is `None`, implying an
    # automatic search for the highest y value in the supplied var_x data.

    # plot`: (*logical*) Do not plot the data on a figure if `False`. Default is
    # `True`.

    # new_window`: (*logical*) Create a new plotting window if `True`. Default is
    # `False`.

    # update`: (*integer*) Update the values of the current snapshot by reading a
    # new simulation output, specified by the output number `update=new_output_number`.
    # This saves the user from having to first run `mydata.update_values(new_output_number)`
    # before calling `plot_histogram` again, it can all be done in one line. Default
    # is `None`.

    # cbar`: (*logical*) Make colorbar next to the plot if `True`. Default is `True`.

    # outline`: (*logical*) Print a contour as an outline around all the data points
    # in the histogram. Default is `False`.

    # summed`: (*logical*) If `True`, the data are summed in the histogram pixels,
    # rather than averaged. If no variable for colour is supplied for either
    # scalar, scatter, contour or image, the colour is used for the number of cells in
    # each pixels and `summed=True` is used. Otherwise, default is `False`.

    # clear`: (*logical*) If `True` the figure is cleared. Default is `True`.

    # block`: (*logical*) If `True` plotting the figure holds the terminal console
    # until it is closed. Default is `False`.

    # equal_axes`: (*logical*) If `True` the aspect ratio is conserved between x and y
    # axes. Default is `False`.

    # scalar_args`: (*dict*) A python dictionary to hold additional matplotlib arguments
    # to be passed to the `contourf` function. Default is empty.

    # image_args`: (*dict*) A python dictionary to hold additional matplotlib arguments
    # to be passed to the `imshow` function. Default is empty.

    # contour_args`: (*dict*) A python dictionary to hold additional matplotlib arguments
    # to be passed to the `contour` function. Default is empty.

    # scatter_args`: (*dict*) A python dictionary to hold additional matplotlib arguments
    # to be passed to the `scatter` function. Default is empty.

    # outline_args`: (*dict*) A python dictionary to hold additional matplotlib arguments
    # to be passed to the `contour` function. Default is empty.


    # Find parent container of object to plot
    holder = var_x.parent

    # Possibility of updating the data from inside the plotting routines
    try:
        update += 0
        holder.update_values(nout=update)
    except TypeError:
        pass

    # Parameters
    nx = resolution+1
    ny = resolution+1

    # Get the data values and units
    datax = holder.get(var_x.name, only_leafs=only_leafs)
    datay = holder.get(var_y.name, only_leafs=only_leafs)
    xlabel = var_x.label+" ["+var_x.unit+"]"
    ylabel = var_y.label+" ["+var_y.unit+"]"
    default_var = "histo_cell_density"

    # Define plotting range
    autoxmin = False
    autoxmax = False
    autoymin = False
    autoymax = False
    try:
        xmin += 0
    except TypeError:
        datax[np.isneginf(datax)] = np.inf
        xmin = np.nanmin(datax)
        autoxmin = True
    try:
        xmax += 0
    except TypeError:
        datax[np.isinf(datax)] = np.NINF
        xmax = np.nanmax(datax)
        autoxmax = True
    try:
        ymin += 0
    except TypeError:
        datay[np.isneginf(datay)] = np.inf
        ymin = np.nanmin(datay)
        autoymin = True
    try:
        ymax += 0
    except TypeError:
        datay[np.isinf(datay)] = np.NINF
        ymax = np.nanmax(datay)
        autoymax = True

    # Protect against empty plots if xmin==xmax or ymin==ymax
    if xmin == xmax:
        if xmin == 0.0:
            xmin = -0.1
            xmax = 0.1
        else:
            xmin = xmin - 0.05*abs(xmin)
            xmax = xmax + 0.05*abs(xmax)
    if ymin == ymax:
        if ymin == 0.0:
            ymin = -0.1
            ymax = 0.1
        else:
            ymin = ymin - 0.05*abs(ymin)
            ymax = ymax + 0.05*abs(ymax)

    dx = xmax-xmin
    dy = ymax-ymin
    if autoxmin:
        xmin = xmin - 0.05*dx
    if autoxmax:
        xmax = xmax + 0.05*dx
    if autoymin:
        ymin = ymin - 0.05*dy
    if autoymax:
        ymax = ymax + 0.05*dy

    # Construct some edge specifiers for the histogram2d function call
    xe = np.linspace(xmin, xmax, nx)
    ye = np.linspace(ymin, ymax, ny)
    # Call the numpy histogram2d function
    z0, yedges1, xedges1 = np.histogram2d(datay, datax, bins=(ye, xe))
    # In the contour plots, x and y are the centers of the cells, instead of the edges.
    x = np.zeros([nx-1])
    y = np.zeros([ny-1])
    for i in range(nx-1):
        x[i] = 0.5*(xe[i]+xe[i+1])
    for j in range(ny-1):
        y[j] = 0.5*(ye[j]+ye[j+1])

    # Use numpy histogram2d function to make image
    z_scal = z_imag = z_cont = z_outl = False
    empty = True
    if scalar:
        try:
            z1, yedges1, xedges1 = np.histogram2d(datay, datax, bins=(
                ye, xe), weights=holder.get(scalar.name, only_leafs=only_leafs))
            if summed:
                z_scal = np.ma.masked_where(z0 == 0.0, z1)
            else:
                with np.errstate(divide="ignore", invalid="ignore"):
                    z_scal = np.ma.masked_where(z0 == 0.0, z1/z0)
        except AttributeError:
            holder.new_field(name=default_var, unit="",
                             label="Number of cells", verbose=True)
            scalar = getattr(holder, default_var)
            z_scal = np.ma.masked_where(z0 == 0.0, z0)
        empty = False
    if image:
        try:
            z1, yedges1, xedges1 = np.histogram2d(datay, datax, bins=(
                ye, xe), weights=holder.get(image.name, only_leafs=only_leafs))
            if summed:
                z_imag = np.ma.masked_where(z0 == 0.0, z1)
            else:
                with np.errstate(divide="ignore", invalid="ignore"):
                    z_imag = np.ma.masked_where(z0 == 0.0, z1/z0)
        except AttributeError:
            holder.new_field(name=default_var, unit="",
                             label="Number of cells", verbose=True)
            image = getattr(holder, default_var)
            z_imag = np.ma.masked_where(z0 == 0.0, z0)
        empty = False
    if contour:
        try:
            z1, yedges1, xedges1 = np.histogram2d(datay, datax, bins=(
                ye, xe), weights=holder.get(contour.name, only_leafs=only_leafs))
            if summed:
                z_cont = np.ma.masked_where(z0 == 0.0, z1)
            else:
                with np.errstate(divide="ignore", invalid="ignore"):
                    z_cont = np.ma.masked_where(z0 == 0.0, z1/z0)
        except AttributeError:
            holder.new_field(name=default_var, unit="",
                             label="Number of cells", verbose=True)
            contour = getattr(holder, default_var)
            z_cont = np.ma.masked_where(z0 == 0.0, z0)
        empty = False
    if scatter:
        empty = False

    # If no variable is requested for z/color dimension, store number of cells by default
    if empty:
        holder.new_field(name=default_var, unit="",
                         label="Number of cells", verbose=True)
        scalar = getattr(holder, default_var)
        z_scal = np.ma.masked_where(z0 == 0.0, z0)

    if outline:
        z_outl = z0

    if plot:
        render_map(scalar=scalar, image=image, contour=contour, scatter=scatter, x=x, y=y, z_scal=z_scal,
                   z_imag=z_imag, z_cont=z_cont, z_outl=z_outl, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, fname=fname,
                   axes=axes, title=title, new_window=new_window, clear=clear, block=block,
                   resolution=resolution, scalar_args=scalar_args, image_args=image_args, holder=holder,
                   contour_args=contour_args, scatter_args=scatter_args, equal_axes=equal_axes, x_raw=datax, y_raw=datay,
                   outline=outline, outline_args=outline_args, sinks=False, only_leafs=only_leafs,
                   dir_vecs=[["", [0, 0, 0]], [var_x.name, [0, 0, 0]], [var_y.name, [0, 0, 0]]])

    if hasattr(holder, default_var):
        holder.delete_field(default_var)

    if copy:
        return x, y, z_scal, z_imag, z_cont, z_outl
    else:
        return

# =======================================================================================
# Plot a 2D slice through the data cube. The arguments are:
# - scalar     : the scalar field to be plotted, e.g. mydata.density
# - image      : the scalar field to be plotted with an image
# - contour    : the scalar field to be plotted with contours
# - vec        : the vector field to be overplotted, e.g. mydata.velocity
# - stream     : the field for streamlines to be overplotted, e.g. mydata.B
# - direction  : the direction normal to the plane of the slice
# - dx         : the x extent of the slice, in units of scale (see data loader)
# - dy         : the y extent of the slice, in units of scale. If not specified, dy = dx
# - dz         : the thickness of the slice
# - axes       : if specified, the data is plotted on the specified axes (see demo).
# - resolution : number of pixels in the slice.
# - fname      : if specified, the figure is saved to file.
# =======================================================================================


def plot_slice(scalar=False, image=False, contour=False, vec=False, stream=False, axes=None,
               direction="z", dx=0.0, dy=0.0, fname=None, title=None, sinks=True, copy=False,
               origin=[0, 0, 0], resolution=128, new_window=False, update=None,
               clear=True, plot=True, block=False, interpolation="linear", sink_args={},
               scalar_args={}, image_args={}, contour_args={}, vec_args={}, stream_args={},
               outline=False, outline_args={}, lmax=0, slice_direction=None):

    # Find parent container of object to plot
    if scalar:
        holder = scalar.parent
    elif image:
        holder = image.parent
    elif contour:
        holder = contour.parent
    elif vec:
        holder = vec.parent
    elif stream:
        holder = stream.parent
    else:
        print("Nothing to plot.")
        return

    if holder.info["ndim"] < 2:
        print("plot_slice Error: Cannot plot slice from 1D data. Exiting...")
        return

    # Possibility of updating the data from inside the plotting routines
    try:
        update += 0
        holder.update_values(nout=update)
    except TypeError:
        pass

    # Determine if interpolation is to be done in 2d or 3d
    if (interpolation.count("3d") > 0) or (interpolation.count("3D") > 0):
        inter_3d = True
        size_fact = np.sqrt(3.0)
        chars = [",", ":", ";", " ", "3d", "3D"]
        for ch in chars:
            interpolation = interpolation.replace(ch, "")
        if len(interpolation) == 0:
            interpolation = "linear"
    else:
        inter_3d = False
        size_fact = 1.0

    # Get slice extent and direction vectors
    if slice_direction is not None:
        [dx, dy, box, dir_vecs, origin] = slice_direction
    else:
        dx, dy, box, dir_vecs, origin = get_slice_direction(
            holder, direction, dx, dy, origin=origin)

    # Try to automatically determine lmax to speedup process
    if lmax == 0:
        dxlmax = holder.info["boxsize_scaled"] * \
            (0.5**holder.info["levelmax_active"])
        target = min(dx, dy)/float(resolution)
        lmax = round(np.log((min(dx, dy)/float(resolution)) /
                            holder.info["boxsize_scaled"])/(np.log(0.5)))
    subset = np.where(np.logical_or(np.logical_and(holder.get("level", only_leafs=False) < lmax,
                                                   holder.get("leaf", only_leafs=False) > 0.0), holder.get("level", only_leafs=False) == lmax))

    # Define equation of a plane
    a_plane = dir_vecs[0][1][0]
    b_plane = dir_vecs[0][1][1]
    c_plane = dir_vecs[0][1][2]
    d_plane = -dir_vecs[0][1][0]*origin[0] - \
        dir_vecs[0][1][1]*origin[1]-dir_vecs[0][1][2]*origin[2]

    # Distance to the plane
    dist1 = (a_plane*holder.get("x", only_leafs=False)[subset] +
             b_plane*holder.get("y", only_leafs=False)[subset] +
             c_plane*holder.get("z", only_leafs=False)[subset] +
             d_plane) / np.sqrt(a_plane**2 + b_plane**2 + c_plane**2)
    # Distance from center
    dist2 = np.sqrt((holder.get("x", only_leafs=False)[subset]-origin[0])**2 +
                    (holder.get("y", only_leafs=False)[subset]-origin[1])**2 +
                    (holder.get("z", only_leafs=False)[subset]-origin[2])**2) - \
        np.sqrt(3.0)*0.5*holder.get("dx", only_leafs=False)[subset]

    # Select only the cells in contact with the slice., i.e. at a distance less than dx/2
    cube = np.where(np.logical_and(np.abs(dist1) <= 0.5*holder.get("dx", only_leafs=False)[subset]*size_fact,
                                   np.abs(dist2) <= max(dx, dy)*0.5*np.sqrt(2.0)))

    # Project coordinates onto the plane by taking dot product with axes vectors
    coords = np.transpose([holder.get("x", only_leafs=False)[subset][cube]-origin[0],
                           holder.get("y", only_leafs=False)[
        subset][cube]-origin[1],
        holder.get("z", only_leafs=False)[subset][cube]-origin[2]])
    datax = np.inner(coords, dir_vecs[1][1])
    datay = np.inner(coords, dir_vecs[2][1])

    # Define slice extent and resolution
    # xmin = max(-0.5*dx, box[0])
    # xmax = min(xmin+dx, box[1])
    # ymin = max(-0.5*dy, box[2])
    # ymax = min(ymin+dy, box[3])
    xmin = -0.5 * dx
    xmax = xmin + dx
    ymin = -0.5 * dy
    ymax = ymin + dy
    nx = resolution
    ny = resolution
    dpx = (xmax-xmin)/float(nx)
    dpy = (ymax-ymin)/float(ny)
    x = np.linspace(xmin+0.5*dpx, xmax-0.5*dpx, nx)
    y = np.linspace(ymin+0.5*dpy, ymax-0.5*dpy, ny)
    grid_x, grid_y = np.meshgrid(x, y)

    # Doing different things depending on whether interpolation is 2D or 3D
    if inter_3d:
        dataz = np.inner(coords, dir_vecs[0][1])
        grid_z = np.zeros([nx, ny])
        points = np.transpose([datax, datay, dataz])
        grids = (grid_x, grid_y, grid_z)
    else:
        points = np.transpose([datax, datay])
        grids = (grid_x, grid_y)

    # Use scipy interpolation function to make image
    z_scal = z_imag = z_cont = u_vect = v_vect = w_vect = u_strm = v_strm = w_strm = 0
    if scalar:
        z_scal = griddata(
            points, scalar.values[subset][cube], grids, method=interpolation)
    if image:
        z_imag = griddata(
            points, image.values[subset][cube], grids, method=interpolation)
    if contour:
        z_cont = griddata(
            points, contour.values[subset][cube], grids, method=interpolation)
    if vec:
        if holder.info["ndim"] < 3:
            datau1 = vec.x.values[subset][cube]
            datav1 = vec.y.values[subset][cube]
        else:
            vectors = np.transpose(
                [vec.x.values[subset][cube], vec.y.values[subset][cube], vec.z.values[subset][cube]])
            datau1 = np.inner(vectors, dir_vecs[1][1])
            datav1 = np.inner(vectors, dir_vecs[2][1])
        u_vect = griddata(points, datau1, grids, method=interpolation)
        v_vect = griddata(points, datav1, grids, method=interpolation)
        if "colors" in vec_args.keys():
            w_vect = griddata(
                points, vec_args["colors"].values[subset][cube], grids, method=interpolation)
        else:
            w_vect = griddata(points, np.sqrt(
                datau1**2+datav1**2), grids, method=interpolation)
    if stream:
        if holder.info["ndim"] < 3:
            datau2 = stream.x.values[subset][cube]
            datav2 = stream.y.values[subset][cube]
        else:
            streams = np.transpose(
                [stream.x.values[subset][cube], stream.y.values[subset][cube], stream.z.values[subset][cube]])
            datau2 = np.inner(streams, dir_vecs[1][1])
            datav2 = np.inner(streams, dir_vecs[2][1])
        u_strm = griddata(points, datau2, grids, method=interpolation)
        v_strm = griddata(points, datav2, grids, method=interpolation)
        w_strm = griddata(points, np.sqrt(datau2**2+datav2**2),
                          grids, method=interpolation)

    # Render the map
    if plot:
        render_map(scalar=scalar, image=image, contour=contour, vec=vec, stream=stream, x=x, y=y, z_scal=z_scal,
                   z_imag=z_imag, z_cont=z_cont, u_vect=u_vect, v_vect=v_vect, w_vect=w_vect, u_strm=u_strm,
                   v_strm=v_strm, w_strm=w_strm, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, fname=fname,
                   axes=axes, title=title, sinks=sinks, new_window=new_window, clear=clear, block=block,
                   resolution=resolution, thePlane=[
                       a_plane, b_plane, c_plane, d_plane],
                   origin=origin, dir_vecs=dir_vecs, scalar_args=scalar_args, image_args=image_args,
                   contour_args=contour_args, vec_args=vec_args, stream_args=stream_args, outline=outline,
                   outline_args=outline_args, sink_args=sink_args, x_raw=datax, y_raw=datay, holder=holder)

    if copy:
        return x, y, z_scal, z_imag, z_cont, u_vect, v_vect, w_vect, u_strm, v_strm, w_strm
    else:
        return

# =======================================================================================
# Plot a column density through the data cube. The arguments are:
# - scalar     : the scalar field to be plotted, e.g. mydata.density
# - image      : the scalar field to be plotted with an image
# - contour    : the scalar field to be plotted with contours
# - dx         : the x extent of the slice, in units of scale (see data loader)
# - dy         : the y extent of the slice, in units of scale. If not specified, dy = dx
# - dz         : the thickness of the slice
# - axes       : if specified, the data is plotted on the specified axes (see demo).
# - resolution : number of pixels in the slice.
# - fname      : if specified, the figure is saved to file.
# =======================================================================================


def plot_column_density(scalar=False, image=False, contour=False, vec=False, stream=False,
                        direction="z", dx=0.0, dy=0.0, dz=0.0, fname=None, axes=None, title=None,
                        origin=[0, 0, 0], resolution=128, sinks=True, summed=True, copy=False,
                        new_window=False, update=None, clear=True, plot=True, block=False, nz=0,
                        interpolation="linear", verbose=False, outline=False, outline_args={},
                        scalar_args={}, image_args={}, contour_args={}, vec_args={}, stream_args={},
                        sink_args={}, lmax=0):

    # Find parent container of object to plot
    if scalar:
        holder = scalar.parent
    elif image:
        holder = image.parent
    elif contour:
        holder = contour.parent
    elif vec:
        holder = vec.parent
    elif stream:
        holder = stream.parent
    else:
        print("Nothing to plot.")
        return

    if holder.info["ndim"] < 2:
        print("plot_column_density Error: Cannot plot slice from 1D data. Exiting...")
        return

    # Possibility of updating the data from inside the plotting routines
    try:
        update += 0
        holder.update_values(nout=update)
    except TypeError:
        pass

    # Get direction vectors once and for all for the column_density.
    # This should be computed here and not inside the plot_slice routine as the origin
    # changes along the z direction inside the loop below.
    dx, dy, box, dir_vecs, origin = get_slice_direction(
        holder, direction, dx, dy, origin=origin)

    # Compute domain dimension for integration
    if dz == 0.0:
        dz = max(dx, dy)
    zmin = -0.5*dz
    zmax = 0.5*dz
    nx = resolution
    ny = resolution
    if nz == 0:
        nz = resolution
    dpz = (zmax-zmin)/float(nz)
    z = np.linspace(zmin+0.5*dpz, zmax-0.5*dpz, nz)

    # We now create empty data arrays that will be filled by the cell data
    z_scal = z_imag = z_cont = u_vect = v_vect = w_vect = u_strm = v_strm = w_strm = 0
    if scalar:
        z_scal = np.zeros([ny, nx])
    if image:
        z_imag = np.zeros([ny, nx])
    if contour:
        z_cont = np.zeros([ny, nx])
    if vec:
        u_vect = np.zeros([ny, nx])
        v_vect = np.zeros([ny, nx])
        w_vect = np.zeros([ny, nx])
    if stream:
        u_strm = np.zeros([ny, nx])
        v_strm = np.zeros([ny, nx])
        w_strm = np.zeros([ny, nx])

    # Define equation of a plane
    a_plane = dir_vecs[0][1][0]
    b_plane = dir_vecs[0][1][1]
    c_plane = dir_vecs[0][1][2]
    d_plane = -dir_vecs[0][1][0]*origin[0] - \
        dir_vecs[0][1][1]*origin[1]-dir_vecs[0][1][2]*origin[2]

    iprog = 1
    istep = 10

    # Begin loop over vertical direction, calling plot_slice with plot=False and copy=True
    # to retrieve the slice data and create a sum
    for iz in range(nz):

        # Print progress
        if verbose:
            percentage = int(float(iz)*100.0/float(nz))
            if percentage >= iprog*istep:
                print("Column density: %3i%% done" % percentage)
                iprog += 1

        [x, y, z_scal_slice, z_imag_slice, z_cont_slice, u_vect_slice, v_vect_slice,
            w_vect_slice, u_strm_slice, v_strm_slice, w_strm_slice] = \
            plot_slice(scalar=scalar, image=image, contour=contour, vec=vec, stream=stream,
                       direction=direction, dx=dx, dy=dy, sinks=sinks, copy=True, resolution=resolution,
                       origin=[origin[0], origin[1], origin[2]+z[iz]], plot=False, interpolation=interpolation, lmax=lmax,
                       slice_direction=[dx, dy, box, dir_vecs, [origin[0], origin[1], origin[2]+z[iz]]])

        # Increment the sum
        if scalar:
            z_scal += z_scal_slice
        if image:
            z_imag += z_imag_slice
        if contour:
            z_cont += z_cont_slice
        if vec:
            u_vect += u_vect_slice
            v_vect += v_vect_slice
            w_vect += w_vect_slice
        if stream:
            u_strm += u_strm_slice
            v_strm += v_strm_slice
            w_strm += w_strm_slice

    # If summed=True, this is a real column density.
    # Else, only the average of the quantity is requested
    if summed:
        column = (zmax-zmin)*conf.constants[holder.info["scale"]]/float(nz)
    else:
        column = 1.0/float(nz)

    if scalar:
        z_scal *= column
    if image:
        z_imag *= column
    if contour:
        z_cont *= column
    if vec:
        u_vect *= column
        v_vect *= column
        w_vect *= column
    if stream:
        u_strm *= column
        v_strm *= column
        w_strm *= column

    # Render the map
    if plot:
        dpx = x[1] - x[0]
        dpy = y[1] - y[0]
        xmin = x[0] - 0.5*dpx
        xmax = x[-1] + 0.5*dpx
        ymin = y[0] - 0.5*dpy
        ymax = y[-1] + 0.5*dpy
        render_map(scalar=scalar, image=image, contour=contour, vec=vec, stream=stream, x=x, y=y, z_scal=z_scal,
                   z_imag=z_imag, z_cont=z_cont, u_vect=u_vect, v_vect=v_vect, w_vect=w_vect, u_strm=u_strm,
                   v_strm=v_strm, w_strm=w_strm, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, fname=fname, dz=dz,
                   axes=axes, title=title, sinks=sinks, new_window=new_window, clear=clear, block=block,
                   resolution=resolution, thePlane=[
                       a_plane, b_plane, c_plane, d_plane],
                   origin=origin, dir_vecs=dir_vecs, scalar_args=scalar_args, image_args=image_args,
                   contour_args=contour_args, vec_args=vec_args, stream_args=stream_args, outline=outline,
                   outline_args=outline_args, sink_args=sink_args, holder=holder)

    if copy:
        return x, y, z_scal, z_imag, z_cont, u_vect, v_vect, w_vect, u_strm, v_strm, w_strm
    else:
        return

# =======================================================================================
# Compute a vector perpendicular to the input vector
# =======================================================================================


def perpendicular_vector(v):

    # x = y = z = 0 is not an acceptable solution
    if v[0] == v[1] == v[2] == 0:
        raise ValueError("zero-vector")

    if v[2] == 0:
        return [-v[1], v[0], 0]
    else:
        return [1.0, 1.0, -1.0 * (v[0] + v[1]) / v[2]]

# =======================================================================================
# Separate arguments to plotting functions between osyris specific arguments and
# Matplotlib arguments.
# =======================================================================================


def parse_arguments(args, args_osyris, args_plot):

    # Save a copy so we can exclude these parameters specific to osyris from the ones
    # to be sent to the matplotlib routines.
    args_osyris["cb_format"] = None
    ignore = set(args_osyris.keys())
    # Now we go through the arguments taken from the function call - scalar_args - and
    # add them to the osyris arguments.
    for key in args.keys():
        args_osyris[key] = args[key]
    # Define colorbar scaling
    cmap = args_osyris["cmap"]
    # norm = None
    # Default number of contours
    try:
        nc = args_osyris["nc"]
    except KeyError:
        nc = 21
    if args_osyris["vmin"] == args_osyris["vmax"]:
        args_osyris["vmin"] /= 1.1
        args_osyris["vmax"] *= 1.1
    norm = Normalize(vmin=args_osyris["vmin"], vmax=args_osyris["vmax"])
    # Default contour levels
    try:
        clevels = np.linspace(args_osyris["vmin"], args_osyris["vmax"], nc)
    except TypeError:
        clevels = [0.0, 1.0]
    # Perform log normalization if it is required
    try:
        if cmap.startswith("log") or cmap.endswith("log"):
            #cmap_save = cmap
            cmap = cmap.replace("log", "")
            chars = [",", ":", ";", " "]
            for ch in chars:
                cmap = cmap.replace(ch, "")
            if len(cmap) == 0:
                cmap = conf.default_values["colormap"]  # cmap_save
            #norm = LogNorm()
            norm = LogNorm(vmin=args_osyris["vmin"], vmax=args_osyris["vmax"])
            clevels = np.logspace(
                np.log10(args_osyris["vmin"]), np.log10(args_osyris["vmax"]), nc)
            args_osyris["cb_format"] = "%.1e"
    except AttributeError:
        pass

    # We now define default parameters for the matplotlib function
    keylist = args_plot.keys()
    if "levels" in keylist:
        args_plot["levels"] = clevels
    if "cmap" in keylist:
        args_plot["cmap"] = cmap
    if "norm" in keylist:
        args_plot["norm"] = norm
    # Then run through the vec_args, adding them to the plotting arguments, but
    # ignoring all osyris specific arguments
    keys = set(args.keys())
    for key in keys.difference(ignore):
        args_plot[key] = args[key]

    return

# =======================================================================================
# Find direction vectors for slice
# =======================================================================================


def get_slice_direction(holder, direction, dx=0, dy=0, origin=[0, 0, 0]):

    # Make it possible to call with only one size in the arguments
    if dy == 0.0:
        dy = dx

    # Transform origin to coordinates if sink is requested
    try:
        if origin.startswith("sink"):
            isink = np.where(holder.sinks["id"] == int(origin.split(":")[1]))[0][0]
            origin = [holder.sinks["x"][isink], holder.sinks["y"]
                      [isink], holder.sinks["z"][isink]]
    except AttributeError:
        pass

    dir_list = {"x": [1, 0, 0], "y": [0, 1, 0], "z": [0, 0, 1]}
    dir_type = len(np.shape(direction))

    if dir_type == 0:  # This is the case where direction contains just one character "x", "y" or "z"
        # This is the case where direction = "auto"
        if direction.startswith("auto"):
            params = direction.split(":")
            if len(params) == 1:
                view = "top"
            else:
                view = params[1]
            if len(params) < 3:
                sphere_rad = 0.5 * \
                    ((np.nanmax(holder.get("x"))-np.nanmin(holder.get("x")))
                     if dx == 0.0 else dx)
            else:
                sphere_rad = float(params[2])
            x_loc = holder.get("x") - origin[0]
            y_loc = holder.get("y") - origin[1]
            z_loc = holder.get("z") - origin[2]
            r_loc = np.linalg.norm([x_loc, y_loc, z_loc], axis=0)
            # Compute angular momentum vector
            sphere = np.where(r_loc < sphere_rad)
            pos = np.vstack((x_loc[sphere], y_loc[sphere],
                             z_loc[sphere])*holder.get("mass")[sphere]).T
            vel = np.vstack((holder.get("velocity_x")[sphere], holder.get(
                "velocity_y")[sphere], holder.get("velocity_z")[sphere])).T
            #vel    = holder.get("velocity")[sphere]
            AngMom = np.sum(np.cross(pos, vel), axis=0)
            if view == "top":
                dir1 = AngMom
                # [1.0, 1.0, -1.0 * (dir1[0] + dir1[1]) / dir1[2]]
                dir2 = perpendicular_vector(dir1)
                dir3 = np.cross(dir1, dir2)
            elif view == "side":
                # Choose a vector perpendicular to the angular momentum vector
                dir3 = AngMom
                # [1.0, 1.0, -1.0 * (dir3[0] + dir3[1]) / dir3[2]]
                dir1 = perpendicular_vector(dir3)
                dir2 = np.cross(dir1, dir3)
            else:
                print("Unknown view direction")
                return
            norm1 = np.linalg.norm(dir1)
            print("Normal slice vector: [%.5e,%.5e,%.5e]" % (
                dir1[0]/norm1, dir1[1]/norm1, dir1[2]/norm1))
            dir_vecs = [["z", dir1], ["x", dir2], ["y", dir3]]
        elif len(direction) == 3:  # This is the case where direction = "xyz"
            dir_vecs = [[direction[0], dir_list[direction[0]]],
                        [direction[1], dir_list[direction[1]]],
                        [direction[2], dir_list[direction[2]]]]
        elif direction == "x":
            dir_vecs = [["x", dir_list["x"]], [
                "y", dir_list["y"]], ["z", dir_list["z"]]]
        elif direction == "y":
            dir_vecs = [["y", dir_list["y"]], [
                "z", dir_list["z"]], ["x", dir_list["x"]]]
        elif direction == "z":
            dir_vecs = [["z", dir_list["z"]], [
                "x", dir_list["x"]], ["y", dir_list["y"]]]
    # This is the case where direction = [1,1,2] (i.e. is a vector with 3 numbers)
    elif dir_type == 1:
        dir1 = direction
        dir2 = perpendicular_vector(dir1)
        dir3 = np.cross(dir1, dir2).tolist()
        dir_vecs = [["z", dir1], ["x", dir2], ["y", dir3]]
    # This is the case where two vectors are specified: direction = [[1,0,1],[0,1,0]]
    elif dir_type == 2:
        dir_vecs = [["z", direction[0]],
                    ["x", direction[1]],
                    ["y", np.cross(direction[0], direction[1]).tolist()]]
    else:
        print("Bad direction for slice: ", direction)
        return

    boxmin_x = np.nanmin(holder.get(dir_vecs[1][0]))
    boxmax_x = np.nanmax(holder.get(dir_vecs[1][0]))
    boxmin_y = np.nanmin(holder.get(dir_vecs[2][0]))
    boxmax_y = np.nanmax(holder.get(dir_vecs[2][0]))
    if dx+dy == 0.0:
        dx = boxmax_x - boxmin_x
        dy = boxmax_y - boxmin_y
    elif dx == 0.0:
        dx = dy

    for i in range(3):
        dir_vecs[i][1] /= np.linalg.norm(dir_vecs[i][1])

    box = [boxmin_x, boxmax_x, boxmin_y, boxmax_y]

    return dx, dy, box, dir_vecs, origin

# =======================================================================================
# Use matplotlib to plot histogram, slice or column density maps
# =======================================================================================


def render_map(scalar=False, image=False, contour=False, scatter=False, vec=False, stream=False, outline=False, x=0, y=0,
               z_scal=0, z_imag=0, z_cont=0, z_outl=0, u_vect=0, v_vect=0, w_vect=0, u_strm=0, v_strm=0,
               w_strm=0, fname=None, axes=None, title=None, sinks=True, new_window=False,
               clear=True, block=False, xmin=0, xmax=0, ymin=0, ymax=0,
               resolution=128, scalar_args={}, image_args={}, contour_args={}, vec_args={},
               stream_args={}, scatter_args={}, outline_args={}, sink_args={}, dz=0, holder=None,
               thePlane=0, origin=[0, 0, 0], dir_vecs=[["z", [0, 0, 0]], ["x", [0, 0, 0]], ["y", [0, 0, 0]]], x_raw=None, y_raw=None, equal_axes=True,
               only_leafs=True):

    if axes:
        theAxes = axes
    elif new_window:
        plt.figure()
        plt.subplot(111)
        theAxes = plt.gca()
    else:
        if clear:
            plt.clf()
        plt.subplot(111)
        theAxes = plt.gca()

    x += np.inner(origin, dir_vecs[1][1])
    y += np.inner(origin, dir_vecs[2][1])

    # Plot scalar field
    if scalar:

        # Round off AMR levels to integers
        if scalar.label == "level" or scalar.label == "cpu":
            z_scal = np.around(z_scal)
        # Parse scalar plot arguments
        scalar_args_osyris = {"vmin": np.nanmin(z_scal), "vmax": np.nanmax(
            z_scal), "cbar": True, "cbax": None, "cmap": conf.default_values["colormap"], "nc": 21}
        scalar_args_plot = {"levels": 1, "cmap": 1, "norm": 1}
        parse_arguments(scalar_args, scalar_args_osyris, scalar_args_plot)
        contf = theAxes.contourf(x, y, z_scal, **scalar_args_plot)
        if scalar_args_osyris["cbar"]:
            scb = plt.colorbar(
                contf, ax=theAxes, cax=scalar_args_osyris["cbax"], format=scalar_args_osyris["cb_format"])
            scb.ax.set_ylabel(
                scalar.label+(" ["+scalar.unit+"]" if len(scalar.unit) > 0 else ""))
            scb.ax.yaxis.set_label_coords(-1.1, 0.5)

    # Plot image
    if image:

        # Round off AMR levels to integers
        if image.label == "level" or image.label == "cpu":
            z_imag = np.around(z_imag)
        # Here we define a set of default parameters
        image_args_osyris = {"vmin": np.nanmin(z_imag), "vmax": np.nanmax(
            z_imag), "cbar": True, "cbax": None, "cmap": conf.default_values["colormap"], "nc": 21}
        # cmap and norm are just dummy arguments to tell the parsing function that they are required
        image_args_plot = {"cmap": 1, "norm": 1,
                           "interpolation": "none", "origin": "lower"}
        parse_arguments(image_args, image_args_osyris, image_args_plot)
        img = theAxes.imshow(
            z_imag, extent=[xmin, xmax, ymin, ymax], **image_args_plot)
        if image_args_osyris["cbar"]:
            icb = plt.colorbar(
                img, ax=theAxes, cax=image_args_osyris["cbax"], format=image_args_osyris["cb_format"])
            icb.ax.set_ylabel(
                image.label+(" ["+image.unit+"]" if len(image.unit) > 0 else ""))
            icb.ax.yaxis.set_label_coords(-1.1, 0.5)

    # Plot contours
    if contour:

        # Round off AMR levels to integers
        if contour.label == "level" or contour.label == "cpu":
            z_cont = np.around(z_cont)
        # Here we define a set of default parameters
        contour_args_osyris = {"vmin": np.nanmin(z_cont), "vmax": np.nanmax(z_cont), "cbar": False, "cbax": None,
                               "cmap": conf.default_values["colormap"], "nc": 21, "label": False, "fmt": "%1.3f"}
        # levels, cmap and norm are just dummy arguments to tell the parsing function that they are required
        contour_args_plot = {"levels": 1, "cmap": 1,
                             "norm": 1, "zorder": 10, "linestyles": "solid"}
        parse_arguments(contour_args, contour_args_osyris, contour_args_plot)
        cont = theAxes.contour(x, y, z_cont, **contour_args_plot)
        if contour_args_osyris["label"]:
            theAxes.clabel(cont, inline=1, fmt=contour_args_osyris["fmt"])
        if contour_args_osyris["cbar"]:
            ccb = plt.colorbar(
                cont, ax=theAxes, cax=contour_args_osyris["cbax"], format=contour_args_osyris["cb_format"])
            ccb.ax.set_ylabel(
                contour.label+(" ["+contour.unit+"]" if len(contour.unit) > 0 else ""))
            ccb.ax.yaxis.set_label_coords(-1.1, 0.5)

    # Plot scatter points
    if scatter:
        cube = np.where(np.logical_and(x_raw >= xmin, np.logical_and(x_raw <= xmax,
                                                                     np.logical_and(y_raw >= ymin, y_raw <= ymax))))
        try:
            vmin = np.nanmin(holder.get(
                scatter.name, only_leafs=only_leafs)[cube])
            vmax = np.nanmax(holder.get(
                scatter.name, only_leafs=only_leafs)[cube])
            scbar = True
            scmap = conf.default_values["colormap"]
        except AttributeError:
            vmin = vmax = None
            scbar = False
            scmap = None
        scatter_args_osyris = {"iskip": 1, "cmap": scmap, "vmin": vmin,
                               "vmax": vmax, "cbar": scbar, "cbax": None, "nc": 21}
        scatter_args_plot = {"cmap": 1, "marker": ".",
                             "c": "b", "edgecolor": "None", "s": 20, "norm": 1}
        parse_arguments(scatter_args, scatter_args_osyris, scatter_args_plot)
        # Check if a variable is given as a color
        try:
            scatter_args_plot["c"] = holder.get(scatter.name, only_leafs=only_leafs)[
                cube][::scatter_args_osyris["iskip"]]
        except AttributeError:
            pass
        scat = theAxes.scatter(x_raw[cube][::scatter_args_osyris["iskip"]],
                               y_raw[cube][::scatter_args_osyris["iskip"]], **scatter_args_plot)
        if scatter_args_osyris["cbar"]:
            rcb = plt.colorbar(
                scat, ax=theAxes, cax=scatter_args_osyris["cbax"], format=scatter_args_osyris["cb_format"])
            rcb.ax.set_ylabel(
                scatter.label+(" ["+scatter.unit+"]" if len(scatter.unit) > 0 else ""))
            rcb.ax.yaxis.set_label_coords(-1.1, 0.5)

    # Draw outline
    if outline:
        outline_args_plot = {"levels": [np.nanmin(z_outl)], "colors": "grey"}
        for key in outline_args.keys():
            outline_args_plot[key] = outline_args[key]
        outl = theAxes.contour(x, y, z_outl, **outline_args_plot)

    # Plot vector field
    if vec:

        vec_args_osyris = {"vskip": int(0.047*resolution), "vscale": np.nanmax(w_vect), "vsize": 15.0, "vkey": True, "vkey_pos": [0.70, -0.08],
                           "cbar": False, "cbax": None, "vmin": np.nanmin(w_vect), "vmax": np.nanmax(w_vect), "nc": 21, "cmap": None, "colors": None,
                           "normalize_arrows": False}
        vec_args_plot = {"cmap": 1, "pivot": "mid", "color": "w", "norm": None}
        parse_arguments(vec_args, vec_args_osyris, vec_args_plot)
        vskip = vec_args_osyris["vskip"]
        if not "scale" in vec_args_plot.keys():
            vec_args_plot["scale"] = vec_args_osyris["vsize"] * \
                vec_args_osyris["vscale"]
        if vec_args_osyris["normalize_arrows"]:
            arrow_norm = np.sqrt(u_vect**2+v_vect**2)
            u_vect = u_vect/arrow_norm
            v_vect = v_vect/arrow_norm
        if vec_args_osyris["colors"]:
            vect = theAxes.quiver(x[::vskip], y[::vskip], u_vect[::vskip, ::vskip], v_vect[::vskip, ::vskip],
                                  w_vect[::vskip, ::vskip], **vec_args_plot)
            if vec_args_osyris["cbar"]:
                vcb = plt.colorbar(
                    vect, ax=theAxes, cax=vec_args_osyris["cbax"], orientation="horizontal", format=vec_args_osyris["cb_format"])
                vcb.ax.set_xlabel(vec_args_osyris["colors"].label+(
                    " ["+vec_args_osyris["colors"].unit+"]" if len(vec_args_osyris["colors"].unit) > 0 else ""))
        elif vec_args_plot["cmap"]:
            vect = theAxes.quiver(x[::vskip], y[::vskip], u_vect[::vskip, ::vskip], v_vect[::vskip, ::vskip],
                                  w_vect[::vskip, ::vskip], **vec_args_plot)
            if vec_args_osyris["cbar"]:
                vcb = plt.colorbar(
                    vect, ax=theAxes, cax=vec_args_osyris["cbax"], orientation="horizontal", format=vec_args_osyris["cb_format"])
                vcb.ax.set_xlabel(
                    vec.label+(" ["+vec.unit+"]" if len(vec.unit) > 0 else ""))
        else:
            vect = theAxes.quiver(x[::vskip], y[::vskip], u_vect[::vskip, ::vskip], v_vect[::vskip, ::vskip],
                                  **vec_args_plot)

        # Plot the scale of the vectors under the axes
        unit_u = vec.unit
        if vec_args_osyris["vkey"]:
            theAxes.quiverkey(vect, vec_args_osyris["vkey_pos"][0], vec_args_osyris["vkey_pos"][1],
                              vec_args_osyris["vscale"], "%.2f [%s]" % (vec_args_osyris["vscale"],
                                                                        unit_u), labelpos="E", labelcolor="k", coordinates="axes", color="k",
                              zorder=100)

    # Plot streamlines
    if stream:

        # Here we define a set of default parameters
        stream_args_osyris = {"cbar": False, "cbax": None, "sskip": 1, "vmin": np.nanmin(
            w_strm), "vmax": np.nanmax(w_strm), "nc": 21, "cmap": None}
        stream_args_plot = {"cmap": 1, "color": "w", "norm": None}
        parse_arguments(stream_args, stream_args_osyris, stream_args_plot)
        sskip = stream_args_osyris["sskip"]
        if stream_args_plot["cmap"]:
            stream_args_plot["color"] = w_strm[::sskip, ::sskip]
        strm = theAxes.streamplot(
            x[::sskip], y[::sskip], u_strm[::sskip, ::sskip], v_strm[::sskip, ::sskip], **stream_args_plot)
        if stream_args_osyris["cbar"]:
            scb = plt.colorbar(
                strm.lines, ax=theAxes, cax=stream_args_osyris["cbax"], orientation="horizontal", format=stream_args_osyris["cb_format"])
            scb.ax.set_xlabel(
                stream.label+(" ["+stream.unit+"]" if len(stream.unit) > 0 else ""))

    # Plot sink particles
    if holder.info["nsinks"] > 0 and sinks:
        dx = xmax-xmin
        if dz == 0:
            thickness = 0.05*dx
        else:
            thickness = 0.5*dz
        dist = (thePlane[0]*holder.sinks["x"]+thePlane[1]*holder.sinks["y"]+thePlane[2]*holder.sinks["z"]+thePlane[3]) \
            / np.sqrt(thePlane[0]**2 + thePlane[1]**2 + thePlane[2]**2)
        sinkcoords = np.transpose(
            [holder.sinks["x"], holder.sinks["y"], holder.sinks["z"]])
        sink_x = np.inner(sinkcoords, dir_vecs[1][1])
        sink_y = np.inner(sinkcoords, dir_vecs[2][1])
        subset = np.where(np.logical_and(dist <= thickness, np.logical_and(
            np.absolute(sink_x) <= 0.5*dx, np.absolute(sink_y) <= 0.5*dx)))
        srad = np.maximum(holder.sinks["radius"]
                          [subset], np.full(len(subset), dx*0.01))
        xy = np.array([sink_x[subset], sink_y[subset]]).T
        patches = [plt.Circle(cent, size) for cent, size in zip(xy, srad)]

        if "colors" in sink_args.keys():
            try:
                # This is a custom array of numbers
                sink_colors = np.array(sink_args["colors"])[subset] + 0.0
            except (TypeError, IndexError):
                # Go and find values in sinks dict
                sink_colors = holder.sinks[sink_args["colors"]][subset]
            sk_vmin = np.nanmin(sink_colors)
            sk_vmax = np.nanmax(sink_colors)
            sk_cmap = conf.default_values["colormap"]
            sk_cbar = True
        else:
            sk_vmin = 0
            sk_vmax = 1
            sk_cmap = None
            sk_cbar = False
        sink_args_osyris = {"cbar": sk_cbar, "cbax": None, "vmin": sk_vmin,
                            "vmax": sk_vmax, "cmap": sk_cmap, "colors": None}
        sink_args_plot = {"facecolors": "w", "edgecolors": "k",
                          "linewidths": 2, "alpha": 0.7, "cmap": 1, "norm": 1}
        parse_arguments(sink_args, sink_args_osyris, sink_args_plot)
        coll = matplotlib.collections.PatchCollection(
            patches, **sink_args_plot)
        if "colors" in sink_args.keys():
            coll.set_array(np.array(sink_colors))
            coll.set_clim([sink_args_osyris["vmin"], sink_args_osyris["vmax"]])
        theAxes.add_collection(coll)
        if sink_args_osyris["cbar"]:
            skcb = plt.colorbar(coll, ax=theAxes, cax=sink_args_osyris["cbax"])
            skcb.ax.set_ylabel("Sink "+sink_args["colors"])
            skcb.ax.yaxis.set_label_coords(-1.1, 0.5)

    try:
        title += ""
        theAxes.set_title(title)
    except TypeError:
        theAxes.set_title("Time = %.3f %s" % (
            holder.info["time"]/conf.constants[conf.default_values["time_unit"]], conf.default_values["time_unit"]))

    if axes:
        pass
    elif new_window:
        theAxes.set_xlim([xmin, xmax])
        theAxes.set_ylim([ymin, ymax])
    else:
        if clear:
            theAxes.set_xlim([xmin, xmax])
            theAxes.set_ylim([ymin, ymax])
        else:
            theAxes.set_xlim([min(theAxes.get_xlim()[0], xmin),
                              max(theAxes.get_xlim()[1], xmax)])
            theAxes.set_ylim([min(theAxes.get_ylim()[0], ymin),
                              max(theAxes.get_ylim()[1], ymax)])

    # Define axes labels
    xlab = getattr(holder, dir_vecs[1][0]).label
    if len(getattr(holder, dir_vecs[1][0]).unit) > 0:
        xlab += " ["+getattr(holder, dir_vecs[1][0]).unit+"]"
    ylab = getattr(holder, dir_vecs[2][0]).label
    if len(getattr(holder, dir_vecs[2][0]).unit) > 0:
        ylab += " ["+getattr(holder, dir_vecs[2][0]).unit+"]"
    theAxes.set_xlabel(xlab)
    theAxes.set_ylabel(ylab)

    if equal_axes:
        theAxes.set_aspect("equal")

    if fname:
        plt.savefig(fname, bbox_inches="tight")
    elif axes:
        pass
    else:
        plt.show(block=block)

    return



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2019 Osyris contributors (https://github.com/nvaytet/osyris)
# @author Neil Vaytet

import numpy as np

def create_hash_table(holder):
    """
    Create a hash table for all the cells in the domain
    """

    print("Building hash table")
    holder.hash_table = dict()
    for icell in range(holder.info["ncells"]):
        igrid = int(holder.get("x")[icell]/holder.get("dx")[icell])
        jgrid = int(holder.get("y")[icell]/holder.get("dx")[icell])
        kgrid = int(holder.get("z")[icell]/holder.get("dx")[icell])
        theHash = str(igrid)+','+str(jgrid)+','+str(kgrid)+','+str(int(holder.get("level")[icell]))
        holder.hash_table[theHash] = icell

    return


def interpolate(field, points):
    """
    Interpolate data at any given point in the whole 3D domain
    """

    holder = field.parent

    try:
        hashTable = holder.hash_table
    except AttributeError:
        print("A hash table is needed to perform interpolations")
        holder.create_hash_table()

    points[:, 0] = ((points[:, 0] + holder.info["xc"]) /
                    holder.info["boxsize_scaled"])
    points[:, 1] = ((points[:, 1] + holder.info["yc"]) /
                    holder.info["boxsize_scaled"])
    points[:, 2] = ((points[:, 2] + holder.info["zc"]) /
                    holder.info["boxsize_scaled"])

    npoints = np.shape(points)[0]
    ilevl = holder.info["levelmax"]
    values = np.zeros([npoints])
    for ip in range(npoints):
        not_found = True
        loop_count = 0
        while not_found:
            l = max(min(ilevl+((-1)**loop_count) *
                        int((loop_count+1)/2), holder.info["levelmax"]), 0)
            loop_count += 1
            dxcell = 0.5**l
            igrid = int(points[ip, 0]/dxcell)
            jgrid = int(points[ip, 1]/dxcell)
            kgrid = int(points[ip, 2]/dxcell)
            theHash = str(igrid)+','+str(jgrid)+','+str(kgrid)+','+str(l)
            try:
                icell = holder.hash_table[theHash]
                ilevl = l
                not_found = False
            except KeyError:
                pass

        cube = dict()
        dmax = 0.0
        cube[theHash] = dict()
        cube[theHash]["vars"] = holder.get(field.name, only_leafs=True)[
            holder.hash_table[theHash]]
        cube[theHash]["dist"] = np.sqrt((points[ip, 0]-((igrid+0.5)*dxcell))**2 +
                                        (points[ip, 1]-((jgrid+0.5)*dxcell))**2 +
                                        (points[ip, 2]-((kgrid+0.5)*dxcell))**2)
        dmax = max(dmax, cube[theHash]["dist"])
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    if i == j == k == 1:
                        pass
                    else:
                        ii = igrid-1+i
                        jj = jgrid-1+j
                        kk = kgrid-1+k
                        theHash = str(ii)+','+str(jj)+',' + \
                            str(kk)+','+str(ilevl)
                        try:
                            neighbour = holder.hash_table[theHash]
                            cube[theHash] = dict()
                            cube[theHash]["vars"] = holder.get(field.name, only_leafs=True)[
                                holder.hash_table[theHash]]
                            cube[theHash]["dist"] = np.sqrt((points[ip, 0]-((ii+0.5)*dxcell))**2 +
                                                            (points[ip, 1]-((jj+0.5)*dxcell))**2 +
                                                            (points[ip, 2]-((kk+0.5)*dxcell))**2)
                            dmax = max(dmax, cube[theHash]["dist"])
                        except KeyError:
                            theHash = str(2*ii)+','+str(2*jj) + \
                                ','+str(2*kk)+','+str(ilevl+1)
                            try:
                                neighbour = holder.hash_table[theHash]
                                for i1 in range(2):
                                    for j1 in range(2):
                                        for k1 in range(2):
                                            theHash = str(
                                                2*ii+i1)+','+str(2*jj+j1)+','+str(2*kk+k1)+','+str(ilevl+1)
                                            cube[theHash] = dict()
                                            cube[theHash]["vars"] = holder.get(field.name, only_leafs=True)[
                                                holder.hash_table[theHash]]
                                            cube[theHash]["dist"] = np.sqrt((points[ip, 0]-((2*ii+i1+0.5)*dxcell*0.5))**2 +
                                                                            (points[ip, 1]-((2*jj+j1+0.5)*dxcell*0.5))**2 +
                                                                            (points[ip, 2]-((2*kk+k1+0.5)*dxcell*0.5))**2)
                                            dmax = max(
                                                dmax, cube[theHash]["dist"])
                            except KeyError:
                                theHash = str(int(float(
                                    ii)/2.0))+','+str(int(float(jj)/2.0))+','+str(int(float(kk)/2.0))+','+str(ilevl-1)
                                try:
                                    neighbour = holder.hash_table[theHash]
                                    cube[theHash] = dict()
                                    cube[theHash]["vars"] = holder.get(field.name, only_leafs=True)[
                                        holder.hash_table[theHash]]
                                    cube[theHash]["dist"] = np.sqrt((points[ip, 0]-((int(float(ii)/2.0)+0.5)*dxcell*2.0))**2 +
                                                                    (points[ip, 1]-((int(float(jj)/2.0)+0.5)*dxcell*2.0))**2 +
                                                                    (points[ip, 2]-((int(float(kk)/2.0)+0.5)*dxcell*2.0))**2)
                                    dmax = max(dmax, cube[theHash]["dist"])
                                except KeyError:
                                    print("Neighbour not found",
                                          igrid, jgrid, kgrid, i, j, k)

        # Compute inverse distance weighting
        result = 0.0
        weights = 0.0
        for key in cube.keys():
            #w = (0.1-1.0)/dmax * cube[key]["dist"] + 1.0
            w = 1.0 / (np.exp(15.0*(cube[key]["dist"]/dmax-0.5)) + 1.0)
            weights += w
            result += w*cube[key]["vars"]

        values[ip] = result/weights

    return values



import osyris
import numpy as np
import matplotlib.pyplot as plt

def test_demo011():

    # Create figure
    fig = plt.figure(figsize=(20, 5))
    ax1 = fig.add_subplot(131)
    ax2 = fig.add_subplot(132)
    ax3 = fig.add_subplot(133)

    # Load data
    mydata = osyris.RamsesData(nout=71, scale="au",
                               verbose=True)

    conv = (mydata.info["unit_l"] / mydata.info["unit_t"])**2

    mydata.new_field(name="internal_energy",
                     operation="passive_scalar_4*density*" + str(conv),
                     unit="erg/cm3",
                     label="Internal energy")

    mydata.print_info()

    # Read EOS table and plot sound speed histogram
    osyris.ism_physics.get_eos(mydata, fname="SCvH_eos.dat")
    mydata.new_field(name="log_cs", operation="np.log10(cs_eos)",
                     label="log(cs)", unit="cm/s")
    osyris.plot_histogram(mydata.log_rho, mydata.log_cs,
                          scalar_args={"cmap": "log,Greens",
                                       "cbar": False},
                          outline=True, axes=ax1,
                          title="Equation of state")

    # Read opacity table and plot Rosseland mean opacity
    osyris.ism_physics.get_opacities(mydata)
    mydata.new_field(name="log_kr", operation="np.log10(kappa_r)",
                     label="log(Kr)", unit="cm2/g")
    osyris.plot_histogram(mydata.log_T, mydata.log_kr,
                          scalar_args={"cmap": "log,Blues",
                                       "cbar": False},
                          outline=True, axes=ax2,
                          title="Opacities")

    # Read resistivity table and plot Ohmic and Ambipolar
    osyris.ism_physics.get_resistivities(mydata)
    mydata.new_field(name="log_etaO", operation="np.log10(eta_ohm)",
                     label="log(etaO)")
    mydata.new_field(name="log_etaA", operation="np.log10(eta_ad)",
                     label="log(etaA)")
    osyris.plot_histogram(mydata.log_rho, mydata.log_etaO,
                          scalar_args={"cmap": "log,Greys",
                                       "cbar": False},
                          outline=True, axes=ax3, title="")
    osyris.plot_histogram(mydata.log_rho, mydata.log_etaA,
                          scalar_args={"cmap": "log,Reds",
                                       "cbar": False},
                          outline=True, axes=ax3,
                          title="Resistivities")
    ax3.set_ylabel("log(eta) [s]")
    ax3.text(-16.0,0.0, "Ambipolar", va="center", ha="center")
    ax3.text(-12.0,-4.0, "Ohmic", va="center", ha="center")

    fig.savefig("ism_tables.png", bbox_inches="tight")


def test_demo012():

    mydata = osyris.RamsesData(nout=71, center="max:density", scale="au")
    osyris.plot_column_density(scalar=mydata.density,
                               direction="z", vec=mydata.velocity,
                               dx=100, dz=20,
                               scalar_args={"cmap": "log"},
                               nz=5, summed=False, fname="demo012.png")


def test_demo013():

    mydata = osyris.RamsesData(nout=71, center="max:density", scale="au")
    osyris.plot_slice(scalar=mydata.density, direction="z",
                      vec=mydata.velocity, dx=100, origin=[0,0,5],
                      scalar_args={"cmap": "log"}, fname="demo013.png")


def test_demo014():

    mydata = osyris.RamsesData(nout=71, center="max:density", scale="au")
    osyris.plot_histogram(mydata.log_rho, mydata.log_T, mydata.mass,
                          summed=True, scalar_args={"cmap": "magma_r,log"},
                          outline=True, fname="demo014.png")


def test_demo015():

    # Load data
    mydata = osyris.RamsesData(nout=71, center="max:density", scale="au")

    # Create figure
    fig = plt.figure(figsize=(20, 10))
    ax1 = fig.add_subplot(231)
    ax2 = fig.add_subplot(232)
    ax3 = fig.add_subplot(233)
    ax4 = fig.add_subplot(234)
    ax5 = fig.add_subplot(235)
    ax6 = fig.add_subplot(236)

    # Density vs B field with AMR level contours
    osyris.plot_histogram(mydata.log_rho, mydata.log_B, axes=ax1,
                          scalar=True, scalar_args={"cmap": "log,YlGnBu"},
                          contour=mydata.level,
                          contour_args={"fmt": "%i", "label": True,
                                        "colors": "k", "cmap": None,
                                        "levels": range(5,20), "cbar": False})

    # Create new field with log of velocity
    mydata.new_field(name="log_vel",
                     operation="np.log10(np.sqrt("
                               "velocity_x**2+velocity_y**2+velocity_z**2))",
                     unit="cm/s", label="log(Velocity)")

    # Density vs log_vel in scatter mode with a grey outline
    osyris.plot_histogram(mydata.log_rho, mydata.log_vel, axes=ax2,
                          scatter=mydata.log_T, outline=True,
                          scatter_args={"iskip": 100, "cmap": "gnuplot"})

    #x,z density slice with B field streamlines
    osyris.plot_slice(mydata.density, direction="yxz", stream=mydata.B,
                      dx=100, axes=ax3, scalar_args={"cmap": "log"})
    # x,y density slice with velocity vectors in color
    osyris.plot_slice(scalar=mydata.log_rho, direction="z",
                      vec=mydata.velocity, dx=100, axes=ax4,
                      vec_args={"cmap": "seismic", "vskip": 4})
    # x,y temperature slice with velocity vectors
    osyris.plot_slice(mydata.log_T, direction="z", vec=mydata.velocity,
                      dx=100, axes=ax5, scalar_args={"cmap":"hot"},
                      contour=mydata.level,
                      contour_args={"fmt": "%i", "label": True,
                                    "colors": "w", "cmap": None,
                                    "levels": range(9,17)})

    # Now update values with later snapshot
    mydata.update_values(201)
    # Re-plot x,y density slice with velocity vectors
    osyris.plot_slice(mydata.log_rho, direction="auto:top",
                      vec=mydata.velocity, dx=100, axes=ax6)

    fig.savefig("demo015.png", bbox_inches="tight")


def test_demo016():

    mydata = osyris.RamsesData(nout=71, center="max:density", scale="au")
    mydata.new_field(name="vkms", operation="velocity/1.0e5",
                     unit="km/s", label="Velocity")
    osyris.plot_slice(scalar=mydata.log_rho, direction="yxz",
                      vec=mydata.B, dx=100,
                      scalar_args={"cmap": "Blues"},
                      vec_args={"cmap":"YlOrRd", "colors": mydata.vkms,
                                "normalize_arrows": True, "vkey": False,
                                "scale": 25.0, "cbar": True, "width": 0.01,
                                "headwidth": 1, "headlength":0},
                      fname="demo016.png")


def test_demo017():

    # Load data
    mydata = osyris.RamsesData(nout=71, center="max:density", scale="au")

    # Create figure
    fig = plt.figure()
    ax = fig.add_subplot(111)

    # Make scatter plot as radial profile
    osyris.plot_histogram(mydata.log_r, mydata.log_rho, scatter=True,
                          scatter_args={"iskip": 100, "c": "grey"},
                          axes=ax)

    # Now overlay mean profile

    # Define min and max range
    rmin = -1.0
    rmax = 4.0

    # Number of points
    nr = 200

    # Radial bin edges and centers
    re = np.linspace(rmin,rmax,nr+1)
    log_r = np.zeros([nr])
    for i in range(nr):
        log_r[i] = 0.5*(re[i]+re[i+1])

    # Modify r values so that the central cell is not "-inf"
    r = np.where(np.isinf(mydata.log_r.values),-2.0,mydata.log_r.values)

    # Bin the data in radial bins
    z0, edges = np.histogram(r, bins=re)
    z1, edges = np.histogram(r, bins=re, weights=mydata.density.values)
    rho_mean = np.log10(z1 / z0)

    # Overlay profile
    ax.plot(log_r, rho_mean, color="r", lw=3, label="Mean profile")
    ax.legend()

    fig.savefig("demo017.png", bbox_inches='tight')


if __name__ == '__main__':
    test_demo017()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2019 Osyris contributors (https://github.com/nvaytet/osyris)
# @author Neil Vaytet

from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import numpy as np

#===================================================================================
# Define default values so that you don't have to specify them every time.
#===================================================================================
default_values = {
    "nout"        : 1    ,
    "lmax"        : 0    ,
    "center"      : None ,
    "dx"          : 0.0  ,
    "dy"          : 0.0  ,
    "dz"          : 0.0  ,
    "scale"       : "au" ,
    "time_unit"   : "kyr",
    "verbose"     : False,
    "path"        : ""   ,
    "variables"   : []   ,
    "colormap"    : "viridis"
}

#===================================================================================
# Define default units for some selected variables. The syntax is:
# "my_variable_name" : ["combination of unit_d, unit_l, unit_t" , "string to be displayed on grid axes"]
#===================================================================================
default_units = {
    ## Example for internal energy
    #"passive_scalar_4" : ["unit_d*((unit_l/unit_t)**2)" , "erg/cm3"],
}

#=======================================================================================
# Common variables
#=======================================================================================
constants = {
    "cm"  : 1.0,
    "au"  : 1.495980e+13,
    "pc"  : 3.085678e+18,
    "s"   : 1.0,
    "yr"  : 365.25*86400.0,
    "kyr" : 365.25*86400.0*1000.0,
    "msun": 1.9889e33,
    "a_r" : 7.56591469318689378e-015,
    "c"   : 2.9979250e+10
}

#=======================================================================================
# Custom colormaps
#=======================================================================================
cmaps = {
    "osyris"  : ["#2b3c4e","#249593","#db6a6c","#ffffff"],
    "osyris2" : ["#2b3c4e","#249593","#ffffff","#db6a6c","#9e4d4e"],
    "osyris3" : ["#3d3d6b","#2a7b9b","#00baad","#57c785","#add45c","#ffc300","#ff8d1a","#ff5733","#c70039","#900c3f","#511849"],
    "osyris4" : ["#000000","#ff5b00","#ffff00","#00ff00","#2bc184","#3d3d6b","#ffffff","#0000ff"],
    "osyris5" : ["#ff0000","#ffff00","#00ff00","#00ffff","#0000ff","#000000","#ffffff"]
}

for key in cmaps.keys():
    cmap = LinearSegmentedColormap.from_list(key, cmaps[key])
    cmap_r = LinearSegmentedColormap.from_list(key+"_r", cmaps[key][::-1])
    plt.register_cmap(cmap=cmap)
    plt.register_cmap(cmap=cmap_r)

#===================================================================================
# Here are some additional variables that are to be computed every time data is
# loaded.
#===================================================================================
def additional_variables(holder):

    # Velocity field (in case conservative variables are dumped)
    holder.new_field(name="velocity",operation="momentum/density",unit="cm/s",label="velocity",verbose=False)

    # Magnetic field
    holder.new_field(name="B",operation="0.5*(B_left+B_right)",unit="G",label="B",verbose=False)

    # Mass and radius
    holder.new_field(name="r",operation="np.sqrt(x**2 + y**2 + z**2)",unit=holder.info["scale"],label="Radius",verbose=False)
    holder.new_field(name="mass",operation="density*((dx*"+str(constants[holder.info["scale"]])+")**3)/"+str(constants["msun"]),unit="Msun",label="Mass",verbose=False)

    # Commonly used log quantities
    holder.new_field(name="log_rho",operation="np.log10(density)",unit="g/cm3",label="log(Density)",verbose=False)
    holder.new_field(name="log_T",operation="np.log10(temperature)",unit="K",label="log(T)",verbose=False)
    holder.new_field(name="log_B",operation="np.log10(B.values)",unit="G",label="log(B)",verbose=False)
    holder.new_field(name="log_m",operation="np.log10(mass)",unit="Msun",label="log(Mass)",verbose=False)
    with np.errstate(divide="ignore"):
        holder.new_field(name="log_r",operation="np.log10(r)",unit=holder.info["scale"],label="log(Radius)",verbose=False)

    # Photon density (in case conservative variables are not dumped)
    if hasattr(holder,"photon_flux_density_1"):
        for igrp in range(holder.info_rt["nGroups"]):
            holder.new_field(name="photon_density_"+str(igrp+1),operation="photon_flux_density_"+str(igrp+1)+"/("+str(constants["c"]*holder.info_rt["rt_c_frac"])+")",unit="photons/cm3",label="photon density "+str(igrp+1),verbose=False)

    #========================== ADD YOUR VARIABLES HERE ============================

    return



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2019 Osyris contributors (https://github.com/nvaytet/osyris)
# @author Neil Vaytet

import numpy as np
import struct
from . import config as conf

#=======================================================================================
# This is the class which will holds a scalar or vector field.
#=======================================================================================
class OsyrisField():

    def __init__(self,values=None,unit=None,label=None,operation=None,depth=None,norm=1.0,\
                 parent=None,kind="scalar",vec_x=False,vec_y=False,vec_z=False,name="",\
                 vector_component=False,group="hydro"):

        self.values = values
        self.unit = unit
        self.label = label
        self.operation = operation
        self.depth = depth
        self.norm = norm
        self.kind = kind
        self.parent = parent
        self.name = name
        self.group = group
        if vec_x:
            self.x = vec_x
        if vec_y:
            self.y = vec_y
        if vec_z:
            self.z = vec_z
        self.vector_component = vector_component

        return

    def get(self, only_leafs=True):
        return self.parent.get(self.name, only_leafs=only_leafs)

#=======================================================================================
# This is a dummy class which gives access to common functions to the other
# classes through inheritance.
#=======================================================================================
class OsyrisData:

    def __init__(self):

        return

    #=======================================================================================
    # Read a test file containing information and parameters
    #=======================================================================================
    def read_parameter_file(self,fname="",dict_name="",evaluate=True,verbose=False,delimiter="="):

        # Read info file and create dictionary
        try:
            with open(fname) as f:
                content = f.readlines()
            f.close()
        except IOError:
            # Clean exit if the file was not found
            if verbose:
                print("File not found: "+fname)
            #if raise_error:
                #raise IOError
            #else:
            return 0

        setattr(self,dict_name,dict())
        for line in content:
            sp = line.split(delimiter)
            if len(sp) > 1:
                if evaluate:
                    try:
                        getattr(self,dict_name)[sp[0].strip()] = eval(sp[1].strip())
                    except (NameError,SyntaxError):
                        getattr(self,dict_name)[sp[0].strip()] = sp[1].strip()
                else:
                    getattr(self,dict_name)[sp[0].strip()] = sp[1].strip()

        return 1

    #=======================================================================================
    # Print information about the data that was loaded.
    #=======================================================================================
    def print_info(self):

        # First get maximum length
        maxlen1 = maxlen2 = maxlen3 = maxlen4 = maxlen5 = maxlen6 = 0
        print_list = dict()
        for key in sorted(self.get_var_list()):
            if not getattr(self,key).vector_component:
                print_list[key] = []
                print_list[key].append(key)
                maxlen1 = max(maxlen1,len(key))
                print_list[key].append(getattr(self,key).kind)
                maxlen2 = max(maxlen2,len(print_list[key][1]))
                print_list[key].append(getattr(self,key).group)
                maxlen3 = max(maxlen3,len(print_list[key][2]))
                print_list[key].append(getattr(self,key).unit)
                maxlen4 = max(maxlen4,len(print_list[key][3]))
                print_list[key].append(str(np.nanmin(self.get(key))))
                print_list[key].append(str(np.nanmax(self.get(key))))
                maxlen5 = max(maxlen5,len(print_list[key][4]))
                maxlen6 = max(maxlen6,len(print_list[key][5]))

        # Now print to screen
        rule = "-" * (maxlen1+maxlen2+maxlen3+maxlen4+maxlen5+maxlen6+7)
        print(rule)
        for key in sorted(self.info.keys()):
            theShape = np.shape(self.info[key])
            if len(theShape) > 0:
                try:
                    print(key+": ["+str(self.info[key][0])+" ... "+str(self.info[key][-1])+"]")
                except IndexError:
                    print(key+": "+str(self.info[key]))
            else:
                print(key+": "+str(self.info[key]))
        print(rule)
        print("The variables are:")
        print("Name".ljust(maxlen1)+" Type".ljust(maxlen2)+"  Group".ljust(maxlen3)+\
              " Unit".ljust(maxlen4)+"    Min".ljust(maxlen5)+"     Max".ljust(maxlen6))
        for key in sorted(print_list.keys()):
            print(print_list[key][0].ljust(maxlen1)+" "+print_list[key][1].ljust(maxlen2)+" "+\
                  print_list[key][2].ljust(maxlen3)+" ["+print_list[key][3].ljust(maxlen4)+"] "+\
                  print_list[key][4].ljust(maxlen5)+" "+print_list[key][5].ljust(maxlen6))
        #print(rule)

        return

    #=======================================================================================
    # The find_center function finds the center in the mesh before loading the full data.
    #=======================================================================================
    def find_center(self,dx,dy,dz):

        lc = False
        try: # check if center is defined at all, if not set to (0.5,0.5,0.5)
            lc = len(self.info["center"])
        except TypeError: # No center defined: set to (0.5,0.5,0.5)
            xc = yc = zc = 0.5
        if lc:
            try: # check if center contains numbers
                self.info["center"][0] += 0
                if lc == 3:
                    xc = self.info["center"][0]
                    yc = self.info["center"][1]
                    zc = self.info["center"][2]
                else:
                    print("Bad center format: must have 3 numbers as input.")
                    return
            except TypeError: # if not it should have the format 'sink:1', or 'max:density'
                if self.info["center"].startswith("sink"):
                    isink = np.where(self.sinks["id"] == int(self.info["center"].split(":")[1]))[0][0]
                    xc = self.sinks["x"][isink]/self.info["boxlen"]/self.info["unit_l"]
                    yc = self.sinks["y"][isink]/self.info["boxlen"]/self.info["unit_l"]
                    zc = self.sinks["z"][isink]/self.info["boxlen"]/self.info["unit_l"]
                else:
                    xc = yc = zc = 0.5

        return xc,yc,zc

    #=======================================================================================
    # The re_center function shifts the coordinates axes around a center. If center="auto"
    # then the function find the cell with the highest density.
    #=======================================================================================
    def re_center(self,newcenter=None):

        try: # check if newcenter is defined
            lc = len(newcenter)

            # Find current center
            xc = self.info["xc"] * conf.constants[self.info["scale"]]
            yc = self.info["yc"] * conf.constants[self.info["scale"]]
            zc = self.info["zc"] * conf.constants[self.info["scale"]]

            # Rescale the coordinates
            self.x.values = self.x.values*conf.constants[self.info["scale"]] + xc
            if self.info["ndim"] > 1:
                self.y.values = self.y.values*conf.constants[self.info["scale"]] + yc
            if self.info["ndim"] > 2:
                self.z.values = self.z.values*conf.constants[self.info["scale"]] + zc

            # Re-scale the cell sizes
            self.dx.values = self.dx.values*conf.constants[self.info["scale"]]

            # Re-center sinks
            if self.info["nsinks"] > 0:
                self.sinks["x"     ] = (self.sinks["x"]+self.info["xc"])*conf.constants[self.info["scale"]]
                self.sinks["y"     ] = (self.sinks["y"]+self.info["yc"])*conf.constants[self.info["scale"]]
                self.sinks["z"     ] = (self.sinks["z"]+self.info["zc"])*conf.constants[self.info["scale"]]
                self.sinks["radius"] =  self.sinks["radius"]/self.info["boxsize"]

            # Re-center particles
            if self.info["npart_tot"] > 0:
                self.part_position_x.values = self.part_position_x.values*conf.constants[self.info["scale"]] + xc
                if self.info["ndim"] > 1:
                    self.part_position_y.values = self.part_position_y.values*conf.constants[self.info["scale"]] + yc
                if self.info["ndim"] > 2:
                    self.part_position_z.values = self.part_position_z.values*conf.constants[self.info["scale"]] + zc

            # Store new center in info
            self.info["center"] = newcenter

        except TypeError:
            pass


        try: # check if center is defined at all, if not set to (0.5,0.5,0.5)
            lc = len(self.info["center"])
            try: # check if center contains numbers
                self.info["center"][0] += 0
                if lc == 3:
                    xc = self.info["center"][0]*self.info["boxsize"]
                    yc = self.info["center"][1]*self.info["boxsize"]
                    zc = self.info["center"][2]*self.info["boxsize"]
                else:
                    print("Bad center format: must have 3 numbers as input.")
                    return
            except TypeError: # if not it should have the format 'sink:1', or 'max:density'
                cvar=self.info["center"].split(":")[1]
                if self.info["center"].startswith("sink"):
                    isink = np.where(self.sinks["id"] == int(cvar))[0][0]
                    xc = self.sinks["x"][isink]
                    yc = self.sinks["y"][isink]
                    zc = self.sinks["z"][isink]
                elif self.info["center"].startswith("max"):
                    # cvar=self.info["center"].split(":")[1]
                    maxloc = np.argmax(self.get(cvar))
                    xc = self.get("x")[maxloc]
                    yc = self.get("y")[maxloc]
                    zc = self.get("z")[maxloc]
                elif self.info["center"].startswith("min"):
                    # cvar=self.info["center"].split(":")[1]
                    minloc = np.argmin(self.get(cvar))
                    xc = self.get("x")[minloc]
                    yc = self.get("y")[minloc]
                    zc = self.get("z")[minloc]
                elif self.info["center"].startswith("av"):
                    # cvar=self.info["center"].split(":")[1]
                    [op_parsed,depth,grp,status] = self.parse_operation(cvar,only_leafs=True)
                    select = eval("np.where("+op_parsed+")")
                    xc = np.average(self.get("x")[select])
                    yc = np.average(self.get("y")[select])
                    zc = np.average(self.get("z")[select])
                else:
                    print("Bad center value:"+str(self.info["center"]))
                    return

        except TypeError: # No center defined: set to (0.5,0.5,0.5)
            xc = yc = zc = 0.5*self.info["boxsize"]

        self.x.values = (self.x.values - xc)/conf.constants[self.info["scale"]]
        if self.info["ndim"] > 1:
            self.y.values = (self.y.values - yc)/conf.constants[self.info["scale"]]
        if self.info["ndim"] > 2:
            self.z.values = (self.z.values - zc)/conf.constants[self.info["scale"]]
        self.info["xc"] = xc/conf.constants[self.info["scale"]]
        self.info["yc"] = yc/conf.constants[self.info["scale"]]
        self.info["zc"] = zc/conf.constants[self.info["scale"]]

        # Re-scale the cell and box sizes
        self.dx.values = self.dx.values/conf.constants[self.info["scale"]]
        self.info["boxsize_scaled"] = self.info["boxsize"]/conf.constants[self.info["scale"]]

        # Re-center sinks
        if self.info["nsinks"] > 0:
            self.sinks["x"     ] = self.sinks["x"]/conf.constants[self.info["scale"]]-self.info["xc"]
            self.sinks["y"     ] = self.sinks["y"]/conf.constants[self.info["scale"]]-self.info["yc"]
            self.sinks["z"     ] = self.sinks["z"]/conf.constants[self.info["scale"]]-self.info["zc"]
            self.sinks["radius"] = self.sinks["radius"]*self.info["boxsize"]/conf.constants[self.info["scale"]]

        # Re-center particles
        if self.info["npart_tot"] > 0:
            self.part_position_x.values = (self.part_position_x.values - xc)/conf.constants[self.info["scale"]]
            if self.info["ndim"] > 1:
                self.part_position_y.values = (self.part_position_y.values - yc)/conf.constants[self.info["scale"]]
            if self.info["ndim"] > 2:
                self.part_position_z.values = (self.part_position_z.values - zc)/conf.constants[self.info["scale"]]

        return

    #=======================================================================================
    # The new field function is used to create a new data field. Say you want to take the
    # log of the density. You create a new field by calling:
    # mydata.new_field(name="log_rho",operation="np.log10(density)",unit="g/cm3",label="log(Density)")
    # The operation string is then evaluated using the 'eval' function.
    #=======================================================================================
    def new_field(self,name,operation="",unit="",label="",verbose=True,values=[],norm=1.0,kind="scalar",\
                  vec_x=False,vec_y=False,vec_z=False,update=False,group=""):

        # Case where values are given and no operation is to be computed
        if (len(operation) == 0) and (len(values) > 0):
            new_data = values
            op_parsed = operation
            depth = -1
            if hasattr(self,name):
                if verbose:
                    print("Warning: field "+name+" already exists and will be overwritten.")
                theField = getattr(self,name)
                theField.values = values
                if not update:
                    theField.unit = unit
                    theField.label = label
                    theField.operation = operation
                    theField.depth = depth
                    theField.norm = norm
                    theField.kind = kind
                    theField.parent = parent
                    if vec_x:
                        theField.x = vec_x
                    if vec_y:
                        theField.y = vec_y
                    if vec_z:
                        theField.z = vec_z
            else:
                if len(group) == 0:
                    group = "hydro"
                dataField = OsyrisField(values=new_data,unit=unit,label=label,operation=op_parsed,depth=depth+1,\
                                       norm=norm,kind=kind,parent=self,vec_x=vec_x,vec_y=vec_y,vec_z=vec_z,name=name,group=group)
                setattr(self, name, dataField)

        # Case where operation is required
        elif (len(operation) > 0) and (len(values) == 0):
            [op_parsed,depth,grp,status] = self.parse_operation(operation)
            if len(group) == 0:
                    group = grp
            if status == 2: # Only scalar fields
                try:
                    with np.errstate(divide="ignore", invalid="ignore"):
                        new_data = eval(op_parsed)
                except NameError:
                    if verbose:
                        print("Error parsing operation when trying to create variable: "+name)
                        print("The attempted operation was: "+op_parsed)
                    return
                dataField = OsyrisField(values=new_data,unit=unit,label=label,operation=op_parsed,depth=depth+1,\
                               norm=norm,kind=kind,parent=self,name=name,group=group)
                if hasattr(self,name) and verbose:
                    print("Warning: field "+name+" already exists and will be overwritten.")
                setattr(self, name, dataField)
            elif status == 1: # Dealing with vector fields
                # Dealing with vector fields: first create x,y,z components
                comps = ["_x","_y","_z"]
                for n in range(self.info["ndim"]):
                    [op_parsed,depth,grp,stat_n] = self.parse_operation(operation,suffix=comps[n])
                    if stat_n == 2:
                        try:
                            with np.errstate(divide="ignore", invalid="ignore"):
                                new_data = eval(op_parsed)
                        except NameError:
                            if verbose:
                                print("Error parsing operation when trying to create variable: "+name+comps[n])
                                print("The attempted operation was: "+op_parsed)
                            return
                        dataField = OsyrisField(values=new_data,unit=unit,label=label,operation=op_parsed,depth=depth+1,\
                                       norm=norm,kind=kind,parent=self,name=name+comps[n],group=group)
                        if hasattr(self,name+comps[n]) and verbose:
                            print("Warning: field "+name+comps[n]+" already exists and will be overwritten.")
                        setattr(self, name+comps[n], dataField)
                    else:
                        print("Error: failed to create vector field.")
                        return
                # Dealing with vector fields: then create vector container
                self.vector_field(name=name,label=label)

        # Case where both values and operation are empty
        elif (len(operation) == 0) and (len(values) == 0):
            dataField = OsyrisField(unit=unit,label=label,parent=self,name=name,group=group)
            setattr(self, name, dataField)
        # Case where both values and operation are required
        else:
            print("Both values and operation are defined. Please choose only one.")

        return

    #=======================================================================================
    # Delete a variable field from the memory
    #=======================================================================================
    def delete_field(self,name):

        delattr(self,name)

        return

    #=======================================================================================
    # The operation parser converts an operation string into an expression which contains
    # variables from the data dictionary. If a name from the variable list, e.g. "density",
    # is found in the operation, it is replaced by self.get("density") so that it
    # can be properly evaluated by the 'eval' function in the 'new_field' function.
    #=======================================================================================
    def parse_operation(self,operation,suffix="",only_leafs=False):

        max_depth = 0
        # Add space before and after to make it easier when searching for characters before
        # and after
        expression = " "+operation+" "
        # Sort the list of variable keys in the order of the longest to the shortest.
        # This guards against replacing 'B' inside 'logB' for example.
        key_list = self.get_var_list()
        key_list = sorted(key_list,key=lambda x:len(x),reverse=True)
        # For replacing, we need to create a list of hash keys to replace on instance at a time
        hashkeys  = dict()
        hashcount = 0
        types_found = {"scalar":False,"vector":False,"hydro":False,"amr":False,"grav":False}

        for key in key_list:

            # First look if there are any ".values" in the operation, i.e. vector magnitudes
            keyVal = key+".values"
            if expression.count(keyVal) > 0:
                hashcount += 1
                theHash = "#"+str(hashcount).zfill(5)+"#"
                hashkeys[theHash] = "self."+keyVal
                expression = expression.replace(keyVal,theHash)
                max_depth = max(max_depth,getattr(self,key).depth)
                types_found["scalar"] = True
                types_found[getattr(self,key).group] = True

            # Now search for all instances of individual variables in string
            loop = True
            loc = 0
            while loop:
                loc = expression.find(key,loc)
                if loc == -1:
                    loop = False
                else:
                    # Check character before and after. If they are either a letter or a '_'
                    # then the instance is actually part of another variable or function name.
                    char_before = expression[loc-1]
                    char_after  = expression[loc+len(key)]
                    bad_before = (char_before.isalpha() or (char_before == "_"))
                    bad_after = (char_after.isalpha() or (char_after == "_"))
                    hashcount += 1
                    if (not bad_before) and (not bad_after):
                        theHash = "#"+str(hashcount).zfill(5)+"#"
                        # Store the data key in the hash table:
                        if getattr(self,key).kind == "vector":
                            thisKey = key+suffix
                        else:
                            thisKey = key
                        hashkeys[theHash] = "self.get(\""+thisKey+"\",only_leafs="+str(only_leafs)+")"
                        expression = expression.replace(key,theHash,1)
                        max_depth = max(max_depth,getattr(self,thisKey).depth)
                        types_found[getattr(self,thisKey).kind] = True
                        types_found[getattr(self,thisKey).group] = True
                    else:
                        # Replace anyway to prevent from replacing "x" in "max("
                        theHash = "#"+str(hashcount).zfill(5)+"#"
                        hashkeys[theHash] = key
                        expression = expression.replace(key,theHash,1)
                    loc += 1
        # Now go through all the hashes in the table and build the final expression
        for theHash in hashkeys.keys():
            expression = expression.replace(theHash,hashkeys[theHash])

        # Determine output group
        if types_found["hydro"]:
            group = "hydro"
        elif types_found["grav"]:
            group = "grav"
        elif types_found["amr"]:
            group = "amr"
        else:
            group = "hydro"

        # Determine exit status
        if types_found["vector"]:
            status = 1
        elif types_found["scalar"]:
            status = 2
        else:
            status = 3

        return [expression,max_depth,group,status]

    #=======================================================================================
    # The function get returns the values of the selected variable.
    # By default, it will only return the leaf cells, but you can choose to return
    # all the cells in the tree by using the argument only_leafs=False.
    #=======================================================================================
    def get(self,var,only_leafs=True):

        # Make sure that we don't use the "only_leafs" indices if we are trying to access
        # particle fields
        if only_leafs and (getattr(self,var).group != "part"):
            return getattr(self,var).values[self.info["leafs"]]
        else:
            return getattr(self,var).values

    #=======================================================================================
    # The function returns the list of variables
    #=======================================================================================
    def get_var_list(self,types=False):
        key_list = []
        typ_list = []
        att_list =  dir(self)
        for att in att_list:
            class_name = getattr(self,att).__class__.__name__
            if class_name == 'OsyrisField':
                key_list.append(att)
                typ_list.append(getattr(self,att).kind)
        if types:
            return [key_list,typ_list]
        else:
            return key_list

    #=======================================================================================
    # Create dummy variables containing the components of the vectors
    #=======================================================================================
    def create_vector_containers(self):

        list_vars = self.get_var_list()

        if self.info["ndim"] > 1:
            for i in range(len(list_vars)):
                key = list_vars[i]
                if key.endswith("_x"):
                    rawkey = key[:-2]
                    ok = True
                    try:
                        k1 = len(self.get(rawkey+"_y"))
                    except AttributeError:
                        ok = False
                    if self.info["ndim"] > 2:
                        try:
                            k2 = len(self.get(rawkey+"_z"))
                        except AttributeError:
                            ok = False

                    if ok:
                        self.vector_field(name=rawkey,label=rawkey)

        return

    #=======================================================================================
    # Create vector field
    #=======================================================================================
    def vector_field(self,name="",values_x=None,values_y=None,values_z=None,unit="",label=""):

        if len(np.shape(values_x)) > 0:
            self.new_field(name+"_x",values=values_x,unit=unit,label=name+"_x",verbose=False)
        if len(np.shape(values_y)) > 0:
            self.new_field(name+"_y",values=values_y,unit=unit,label=name+"_y",verbose=False)
        if len(np.shape(values_z)) > 0:
            self.new_field(name+"_z",values=values_z,unit=unit,label=name+"_z",verbose=False)

        v_x=getattr(self,name+"_x")
        v_y=getattr(self,name+"_y")
        v_x.vector_component = True
        v_y.vector_component = True

        if self.info["ndim"] > 2:
            v_z=getattr(self,name+"_z")
            v_z.vector_component = True
            vals = np.linalg.norm([self.get(name+"_x",only_leafs=False),\
                                   self.get(name+"_y",only_leafs=False),\
                                   self.get(name+"_z",only_leafs=False)],axis=0)
        else:
            v_z = False
            vals = np.linalg.norm([self.get(name+"_x",only_leafs=False),self.get(name+"_y",only_leafs=False)],axis=0)

        self.new_field(name=name,values=vals,label=label,vec_x=v_x,vec_y=v_y,vec_z=v_z,kind="vector",unit=v_x.unit,group=v_x.group)

        return

    #=======================================================================================
    # Get cylindrical basis
    #=======================================================================================
    def get_cylindrical_basis(self,direction):

        pos = np.vstack((self.x.values,self.y.values,self.z.values)).T
        ez   = direction/np.linalg.norm(direction)
        ephi = np.cross(ez,pos)
        ephi_norm = np.linalg.norm(ephi,axis=1)
        ephi = np.vstack((ephi[:,0]/ephi_norm,ephi[:,1]/ephi_norm,ephi[:,2]/ephi_norm)).T
        er   = np.cross(ephi,ez)

        del pos
        return er,ephi,ez

    #=======================================================================================
    # Get cylindrical components of the vector field variable
    #=======================================================================================
    def get_cylindrical_components(self,variable,direction):

        if getattr(self,variable).kind != 'vector':
            print("get_cylindrical_components must be applied to a vector field!")
            return

        if hasattr(self,variable+"_cyl_r"):
            print("****** Warning ****** : cylindrical components of "+variable+" already exist")
            print("you should check if the vector direction was the good one and/or delete the previously computed components")
            return

        er,ephi,ez = self.get_cylindrical_basis(direction)
        vec = np.vstack((getattr(self,variable+"_x").values,getattr(self,variable+"_y").values,getattr(self,variable+"_z").values)).T

        self.new_field(name=variable+"_cyl_r"  ,values=np.sum(vec*er  ,axis=1),unit=getattr(self,variable).unit,label=getattr(self,variable).label+"_r")
        self.new_field(name=variable+"_cyl_phi",values=np.sum(vec*ephi,axis=1),unit=getattr(self,variable).unit,label=getattr(self,variable).label+"_phi")
        self.new_field(name=variable+"_cyl_z"  ,values=np.sum(vec*ez  ,axis=1),unit=getattr(self,variable).unit,label=getattr(self,variable).label+"_z")

        del er,ephi,ez,vec
        return

#=======================================================================================
#=======================================================================================
# End of class OsyrisData()
#=======================================================================================
#=======================================================================================


#=======================================================================================
#=======================================================================================
# USEFUL TOOLS
#=======================================================================================
#=======================================================================================

#=======================================================================================
# Determine binary offset when reading fortran binary files and return unpacked data
#=======================================================================================
def get_binary_data(fmt="",ninteg=0,nlines=0,nfloat=0,nstrin=0,nquadr=0,nlongi=0,offset=None,content=None,correction=0):

    if offset is None:
        offset = 4*ninteg + 8*(nlines+nfloat+nlongi) + nstrin + nquadr*16
    offset += 4 + correction
    byte_size = {"b": 1 , "h": 2, "i": 4, "q": 8, "f": 4, "d": 8, "e": 8}
    if len(fmt) == 1:
        mult = 1
    else:
        mult = eval(fmt[0:len(fmt)-1])
    pack_size = mult*byte_size[fmt[-1]]

    return struct.unpack(fmt, content[offset:offset+pack_size])
